from enum import Enum


class StatusKind(str, Enum):
    BLOCKED = "Blocked"
    CANCELED = "Canceled"
    FINISHED = "Finished"
    STARTED = "Started"
    UNSTARTED = "Unstarted"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any, Dict, List, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.color_name import ColorName
from ..models.icon_kind import IconKind
from ..types import UNSET, Unset

T = TypeVar("T", bound="ViewCreate")


@_attrs_define
class ViewCreate:
    """
    Attributes:
        duid (str):
        order (str):
        layout_duid (str):
        accessible_by_team (Union[Unset, bool]):
        accessible_by_user_duids (Union[Unset, List[str]]):
        title (Union[Unset, str]):
        description (Union[Unset, str]):
        icon_kind (Union[Unset, IconKind]): * `None` - NONE
            * `Icon` - ICON
            * `Emoji` - EMOJI
        icon_name_or_emoji (Union[Unset, str]):
        color_name (Union[Unset, ColorName]): * `Red` - RED
            * `Dark Blue` - DARK_BLUE
            * `Dark Orange` - DARK_ORANGE
            * `Dark Green` - DARK_GREEN
            * `Purple` - PURPLE
            * `Dark Teal` - DARK_TEAL
            * `Pink` - PINK
            * `Orange` - ORANGE
            * `Green` - GREEN
            * `Yellow` - YELLOW
            * `Brown` - BROWN
            * `Dark Red` - DARK_RED
            * `Flat Green` - FLAT_GREEN
            * `Red Orange` - RED_ORANGE
            * `Teal` - TEAL
            * `Light Green` - LIGHT_GREEN
            * `Light Blue` - LIGHT_BLUE
            * `Light Purple` - LIGHT_PURPLE
            * `Light Orange` - LIGHT_ORANGE
            * `Light Pink` - LIGHT_PINK
            * `Tan` - TAN
            * `Dark Gray` - DARK_GRAY
            * `Light Brown` - LIGHT_BROWN
            * `Light Gray` - LIGHT_GRAY
        favorited_by_user_duids (Union[Unset, List[str]]):
    """

    duid: str
    order: str
    layout_duid: str
    accessible_by_team: Union[Unset, bool] = UNSET
    accessible_by_user_duids: Union[Unset, List[str]] = UNSET
    title: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    icon_kind: Union[Unset, IconKind] = UNSET
    icon_name_or_emoji: Union[Unset, str] = UNSET
    color_name: Union[Unset, ColorName] = UNSET
    favorited_by_user_duids: Union[Unset, List[str]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        order = self.order
        layout_duid = self.layout_duid
        accessible_by_team = self.accessible_by_team
        accessible_by_user_duids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.accessible_by_user_duids, Unset):
            accessible_by_user_duids = self.accessible_by_user_duids

        title = self.title
        description = self.description
        icon_kind: Union[Unset, str] = UNSET
        if not isinstance(self.icon_kind, Unset):
            icon_kind = self.icon_kind.value

        icon_name_or_emoji = self.icon_name_or_emoji
        color_name: Union[Unset, str] = UNSET
        if not isinstance(self.color_name, Unset):
            color_name = self.color_name.value

        favorited_by_user_duids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.favorited_by_user_duids, Unset):
            favorited_by_user_duids = self.favorited_by_user_duids

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "order": order,
                "layoutDuid": layout_duid,
            }
        )
        if accessible_by_team is not UNSET:
            field_dict["accessibleByTeam"] = accessible_by_team
        if accessible_by_user_duids is not UNSET:
            field_dict["accessibleByUserDuids"] = accessible_by_user_duids
        if title is not UNSET:
            field_dict["title"] = title
        if description is not UNSET:
            field_dict["description"] = description
        if icon_kind is not UNSET:
            field_dict["iconKind"] = icon_kind
        if icon_name_or_emoji is not UNSET:
            field_dict["iconNameOrEmoji"] = icon_name_or_emoji
        if color_name is not UNSET:
            field_dict["colorName"] = color_name
        if favorited_by_user_duids is not UNSET:
            field_dict["favoritedByUserDuids"] = favorited_by_user_duids

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        order = d.pop("order")

        layout_duid = d.pop("layoutDuid")

        accessible_by_team = d.pop("accessibleByTeam", UNSET)

        accessible_by_user_duids = cast(List[str], d.pop("accessibleByUserDuids", UNSET))

        title = d.pop("title", UNSET)

        description = d.pop("description", UNSET)

        _icon_kind = d.pop("iconKind", UNSET)
        icon_kind: Union[Unset, IconKind]
        if isinstance(_icon_kind, Unset):
            icon_kind = UNSET
        else:
            icon_kind = IconKind(_icon_kind)

        icon_name_or_emoji = d.pop("iconNameOrEmoji", UNSET)

        _color_name = d.pop("colorName", UNSET)
        color_name: Union[Unset, ColorName]
        if isinstance(_color_name, Unset):
            color_name = UNSET
        else:
            color_name = ColorName(_color_name)

        favorited_by_user_duids = cast(List[str], d.pop("favoritedByUserDuids", UNSET))

        view_create = cls(
            duid=duid,
            order=order,
            layout_duid=layout_duid,
            accessible_by_team=accessible_by_team,
            accessible_by_user_duids=accessible_by_user_duids,
            title=title,
            description=description,
            icon_kind=icon_kind,
            icon_name_or_emoji=icon_name_or_emoji,
            color_name=color_name,
            favorited_by_user_duids=favorited_by_user_duids,
        )

        view_create.additional_properties = d
        return view_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.color_name import ColorName
from ..models.theme import Theme
from ..types import UNSET, Unset

T = TypeVar("T", bound="UserCreate")


@_attrs_define
class UserCreate:
    """
    Attributes:
        duid (str):
        name (Union[Unset, str]):
        theme (Union[Unset, Theme]): * `System Default` - SYSTEM_DEFAULT
            * `Light` - LIGHT
            * `Dark` - DARK
        color_name (Union[Unset, ColorName]): * `Red` - RED
            * `Dark Blue` - DARK_BLUE
            * `Dark Orange` - DARK_ORANGE
            * `Dark Green` - DARK_GREEN
            * `Purple` - PURPLE
            * `Dark Teal` - DARK_TEAL
            * `Pink` - PINK
            * `Orange` - ORANGE
            * `Green` - GREEN
            * `Yellow` - YELLOW
            * `Brown` - BROWN
            * `Dark Red` - DARK_RED
            * `Flat Green` - FLAT_GREEN
            * `Red Orange` - RED_ORANGE
            * `Teal` - TEAL
            * `Light Green` - LIGHT_GREEN
            * `Light Blue` - LIGHT_BLUE
            * `Light Purple` - LIGHT_PURPLE
            * `Light Orange` - LIGHT_ORANGE
            * `Light Pink` - LIGHT_PINK
            * `Tan` - TAN
            * `Dark Gray` - DARK_GRAY
            * `Light Brown` - LIGHT_BROWN
            * `Light Gray` - LIGHT_GRAY
        open_in_native_app (Union[Unset, bool]):
        notification_default (Union[Unset, bool]):
        notification_email (Union[Unset, bool]):
        notification_slack (Union[Unset, bool]):
    """

    duid: str
    name: Union[Unset, str] = UNSET
    theme: Union[Unset, Theme] = UNSET
    color_name: Union[Unset, ColorName] = UNSET
    open_in_native_app: Union[Unset, bool] = UNSET
    notification_default: Union[Unset, bool] = UNSET
    notification_email: Union[Unset, bool] = UNSET
    notification_slack: Union[Unset, bool] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        name = self.name
        theme: Union[Unset, str] = UNSET
        if not isinstance(self.theme, Unset):
            theme = self.theme.value

        color_name: Union[Unset, str] = UNSET
        if not isinstance(self.color_name, Unset):
            color_name = self.color_name.value

        open_in_native_app = self.open_in_native_app
        notification_default = self.notification_default
        notification_email = self.notification_email
        notification_slack = self.notification_slack

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
            }
        )
        if name is not UNSET:
            field_dict["name"] = name
        if theme is not UNSET:
            field_dict["theme"] = theme
        if color_name is not UNSET:
            field_dict["colorName"] = color_name
        if open_in_native_app is not UNSET:
            field_dict["openInNativeApp"] = open_in_native_app
        if notification_default is not UNSET:
            field_dict["notificationDefault"] = notification_default
        if notification_email is not UNSET:
            field_dict["notificationEmail"] = notification_email
        if notification_slack is not UNSET:
            field_dict["notificationSlack"] = notification_slack

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        name = d.pop("name", UNSET)

        _theme = d.pop("theme", UNSET)
        theme: Union[Unset, Theme]
        if isinstance(_theme, Unset):
            theme = UNSET
        else:
            theme = Theme(_theme)

        _color_name = d.pop("colorName", UNSET)
        color_name: Union[Unset, ColorName]
        if isinstance(_color_name, Unset):
            color_name = UNSET
        else:
            color_name = ColorName(_color_name)

        open_in_native_app = d.pop("openInNativeApp", UNSET)

        notification_default = d.pop("notificationDefault", UNSET)

        notification_email = d.pop("notificationEmail", UNSET)

        notification_slack = d.pop("notificationSlack", UNSET)

        user_create = cls(
            duid=duid,
            name=name,
            theme=theme,
            color_name=color_name,
            open_in_native_app=open_in_native_app,
            notification_default=notification_default,
            notification_email=notification_email,
            notification_slack=notification_slack,
        )

        user_create.additional_properties = d
        return user_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.property_kind import PropertyKind
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.property_create_adtl import PropertyCreateAdtl


T = TypeVar("T", bound="PropertyCreate")


@_attrs_define
class PropertyCreate:
    """
    Attributes:
        duid (str):
        kind (PropertyKind): * `Default: Status` - DEFAULT_STATUS
            * `Default: Assignee` - DEFAULT_ASSIGNEE
            * `Default: Dates` - DEFAULT_DATES
            * `Default: Priority` - DEFAULT_PRIORITY
            * `Default: Tags` - DEFAULT_TAGS
            * `Default: Size` - DEFAULT_SIZE
            * `Default: Created` - DEFAULT_CREATED_AT
            * `Default: Created by` - DEFAULT_CREATED_BY
            * `Default: Last updated` - DEFAULT_UPDATED_AT
            * `Default: Last updated by` - DEFAULT_UPDATED_BY
            * `Text` - TEXT
            * `Number` - NUMBER
            * `Checkbox` - CHECKBOX
            * `Select` - SELECT
            * `Multiselect` - MULTISELECT
            * `Status` - STATUS
            * `User` - USER
            * `Dates` - DATES
        order (str):
        hidden (Union[Unset, bool]):
        title (Union[Unset, str]):
        description (Union[Unset, str]):
        adtl (Union[Unset, PropertyCreateAdtl]):
    """

    duid: str
    kind: PropertyKind
    order: str
    hidden: Union[Unset, bool] = UNSET
    title: Union[Unset, str] = UNSET
    description: Union[Unset, str] = UNSET
    adtl: Union[Unset, "PropertyCreateAdtl"] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        kind = self.kind.value

        order = self.order
        hidden = self.hidden
        title = self.title
        description = self.description
        adtl: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.adtl, Unset):
            adtl = self.adtl.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "kind": kind,
                "order": order,
            }
        )
        if hidden is not UNSET:
            field_dict["hidden"] = hidden
        if title is not UNSET:
            field_dict["title"] = title
        if description is not UNSET:
            field_dict["description"] = description
        if adtl is not UNSET:
            field_dict["adtl"] = adtl

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.property_create_adtl import PropertyCreateAdtl

        d = src_dict.copy()
        duid = d.pop("duid")

        kind = PropertyKind(d.pop("kind"))

        order = d.pop("order")

        hidden = d.pop("hidden", UNSET)

        title = d.pop("title", UNSET)

        description = d.pop("description", UNSET)

        _adtl = d.pop("adtl", UNSET)
        adtl: Union[Unset, PropertyCreateAdtl]
        if isinstance(_adtl, Unset):
            adtl = UNSET
        else:
            adtl = PropertyCreateAdtl.from_dict(_adtl)

        property_create = cls(
            duid=duid,
            kind=kind,
            order=order,
            hidden=hidden,
            title=title,
            description=description,
            adtl=adtl,
        )

        property_create.additional_properties = d
        return property_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Optional, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.relationship_kind_kind import RelationshipKindKind
from ..types import UNSET, Unset

T = TypeVar("T", bound="RelationshipKind")


@_attrs_define
class RelationshipKind:
    """
    Attributes:
        duid (str):
        kind (RelationshipKindKind): * `Parent Of` - PARENT_OF
            * `Blocks` - BLOCKS
            * `Relates To` - RELATES_TO
            * `Duplicates` - DUPLICATES
            * `Custom` - CUSTOM
        directed (bool):
        forward_text (str):
        updated_by_client_duid (Union[Unset, None, str]):
        backward_text (Optional[str]):
    """

    duid: str
    kind: RelationshipKindKind
    directed: bool
    forward_text: str
    backward_text: Optional[str]
    updated_by_client_duid: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        kind = self.kind.value

        directed = self.directed
        forward_text = self.forward_text
        updated_by_client_duid = self.updated_by_client_duid
        backward_text = self.backward_text

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "kind": kind,
                "directed": directed,
                "forwardText": forward_text,
                "backwardText": backward_text,
            }
        )
        if updated_by_client_duid is not UNSET:
            field_dict["updatedByClientDuid"] = updated_by_client_duid

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        kind = RelationshipKindKind(d.pop("kind"))

        directed = d.pop("directed")

        forward_text = d.pop("forwardText")

        updated_by_client_duid = d.pop("updatedByClientDuid", UNSET)

        backward_text = d.pop("backwardText")

        relationship_kind = cls(
            duid=duid,
            kind=kind,
            directed=directed,
            forward_text=forward_text,
            updated_by_client_duid=updated_by_client_duid,
            backward_text=backward_text,
        )

        relationship_kind.additional_properties = d
        return relationship_kind

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ViewKind(str, Enum):
    CUSTOM = "Custom"
    MY_TASKS = "My tasks"
    SEARCH = "Search"
    TRASH = "Trash"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.filter_connector import FilterConnector

if TYPE_CHECKING:
    from ..models.filter_assignee import FilterAssignee
    from ..models.filter_search import FilterSearch
    from ..models.filter_set import FilterSet


T = TypeVar("T", bound="FilterGroup")


@_attrs_define
class FilterGroup:
    """
    Attributes:
        filters (Union['FilterAssignee', 'FilterSearch', 'FilterSet']):
        connector (FilterConnector): * `or` - OR
            * `and` - AND
    """

    filters: Union["FilterAssignee", "FilterSearch", "FilterSet"]
    connector: FilterConnector
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        from ..models.filter_assignee import FilterAssignee
        from ..models.filter_set import FilterSet

        filters: Dict[str, Any]

        if isinstance(self.filters, FilterAssignee):
            filters = self.filters.to_dict()

        elif isinstance(self.filters, FilterSet):
            filters = self.filters.to_dict()

        else:
            filters = self.filters.to_dict()

        connector = self.connector.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "filters": filters,
                "connector": connector,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.filter_assignee import FilterAssignee
        from ..models.filter_search import FilterSearch
        from ..models.filter_set import FilterSet

        d = src_dict.copy()

        def _parse_filters(data: object) -> Union["FilterAssignee", "FilterSearch", "FilterSet"]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                componentsschemas_filter_type_0 = FilterAssignee.from_dict(data)

                return componentsschemas_filter_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                componentsschemas_filter_type_1 = FilterSet.from_dict(data)

                return componentsschemas_filter_type_1
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            componentsschemas_filter_type_2 = FilterSearch.from_dict(data)

            return componentsschemas_filter_type_2

        filters = _parse_filters(d.pop("filters"))

        connector = FilterConnector(d.pop("connector"))

        filter_group = cls(
            filters=filters,
            connector=connector,
        )

        filter_group.additional_properties = d
        return filter_group

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class EventKind(str, Enum):
    AICONTENT = "ai/content"
    AIEMOJI = "ai/emoji"
    AIFEEDBACK = "ai/feedback"
    AIICON = "ai/icon"
    AIPROPS = "ai/props"
    AISUBTASKS = "ai/subtasks"
    AITRANSLATE = "ai/translate"
    DOCSCREATE = "docs/create"
    DOCSDELETE = "docs/delete"
    DOCSUPDATE_OTHER = "docs/update_other"
    DOCSUPDATE_TITLE = "docs/update_title"
    HELPRESOURCE_CLICK = "help/resource_click"
    ONBOARDINGFINISH_STEP = "onboarding/finish_step"
    PAGESCREATE = "pages/create"
    PAGESDELETE = "pages/delete"
    PAGESROLLOVER = "pages/rollover"
    PAGESUPDATE_OTHER = "pages/update_other"
    PAGESUPDATE_PERMISSIONS = "pages/update_permissions"
    PAGESUPDATE_TITLE = "pages/update_title"
    PROFILEBECOME_ACTIVE = "profile/become_active"
    PROFILEBECOME_INACTIVE = "profile/become_inactive"
    PROFILEDELETE = "profile/delete"
    PROFILEUPDATE = "profile/update"
    TASKSCOMMENT = "tasks/comment"
    TASKSCREATE = "tasks/create"
    TASKSDELETE = "tasks/delete"
    TASKSDELETE_FULLY = "tasks/delete_fully"
    TASKSDUE_DATE_TOMORROW = "tasks/due_date_tomorrow"
    TASKSREMINDER_NOW = "tasks/reminder_now"
    TASKSREMOVE_RELATIONSHIP = "tasks/remove_relationship"
    TASKSSET_RELATIONSHIP = "tasks/set_relationship"
    TASKSUPDATE_ASSIGNEES = "tasks/update_assignees"
    TASKSUPDATE_DESCRIPTION = "tasks/update_description"
    TASKSUPDATE_OTHER = "tasks/update_other"
    TASKSUPDATE_STATUS = "tasks/update_status"
    TASKSUPDATE_SUBSCRIPTIONS = "tasks/update_subscriptions"
    TASKSUPDATE_TITLE = "tasks/update_title"
    USAGECOPY_BRANCH = "usage/copy_branch"
    USAGECOPY_TASK_LINK = "usage/copy_task_link"
    USAGENLP_RAW_CREATE = "usage/nlp_raw_create"
    USAGENLP_RAW_DELETE = "usage/nlp_raw_delete"
    USAGENLP_TYPEAHEAD_ACCEPT = "usage/nlp_typeahead_accept"
    USAGENLP_TYPEAHEAD_OPEN = "usage/nlp_typeahead_open"
    USAGEOPEN_COMMAND_CENTER = "usage/open_command_center"
    USAGEOPEN_FULLSCREEN = "usage/open_fullscreen"
    USAGEOPEN_RIGHTBAR = "usage/open_rightbar"
    USAGEOPEN_SEARCH = "usage/open_search"
    USAGEREDO = "usage/redo"
    USAGESUBMIT_FEEDBACK = "usage/submit_feedback"
    USAGEUNDO = "usage/undo"
    WORKSPACEBECOME_ACTIVE = "workspace/become_active"
    WORKSPACEBECOME_INACTIVE = "workspace/become_inactive"
    WORKSPACECREATE = "workspace/create"
    WORKSPACEDATA_EXPORT = "workspace/data_export"
    WORKSPACEDATA_IMPORT = "workspace/data_import"
    WORKSPACEDELETE = "workspace/delete"
    WORKSPACEDOWNGRADE_FINALIZE = "workspace/downgrade_finalize"
    WORKSPACEDOWNGRADE_INITIALIZE = "workspace/downgrade_initialize"
    WORKSPACEINVITE = "workspace/invite"
    WORKSPACEJOIN = "workspace/join"
    WORKSPACELEAVE = "workspace/leave"
    WORKSPACEUPDATE_OTHER = "workspace/update_other"
    WORKSPACEUPDATE_PROPERTY = "workspace/update_property"
    WORKSPACEUPDATE_STATUS = "workspace/update_status"
    WORKSPACEUPGRADE = "workspace/upgrade"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class TransactionKind(str, Enum):
    COMMENT_CREATE = "comment_create"
    COMMENT_DELETE = "comment_delete"
    COMMENT_REACTION_CREATE = "comment_reaction_create"
    COMMENT_REACTION_DELETE = "comment_reaction_delete"
    COMMENT_UPDATE = "comment_update"
    CYCLE_ROLLOVER = "cycle_rollover"
    DARTBOARD_CREATE = "dartboard_create"
    DARTBOARD_DELETE = "dartboard_delete"
    DARTBOARD_UPDATE = "dartboard_update"
    DOC_CREATE = "doc_create"
    DOC_DELETE = "doc_delete"
    DOC_UPDATE = "doc_update"
    EVENT_CREATE = "event_create"
    FOLDER_CREATE = "folder_create"
    FOLDER_DELETE = "folder_delete"
    FOLDER_UPDATE = "folder_update"
    LAYOUT_CREATE = "layout_create"
    LAYOUT_DELETE = "layout_delete"
    LAYOUT_UPDATE = "layout_update"
    OPTION_CREATE = "option_create"
    OPTION_DELETE = "option_delete"
    OPTION_UPDATE = "option_update"
    PROPERTY_CREATE = "property_create"
    PROPERTY_DELETE = "property_delete"
    PROPERTY_UPDATE = "property_update"
    RELATIONSHIP_CREATE = "relationship_create"
    RELATIONSHIP_DELETE = "relationship_delete"
    RELATIONSHIP_UPDATE = "relationship_update"
    SPACE_CREATE = "space_create"
    SPACE_DELETE = "space_delete"
    SPACE_UPDATE_OTHER = "space_update_other"
    SPACE_UPDATE_PERMS = "space_update_perms"
    STATUS_CREATE = "status_create"
    STATUS_DELETE = "status_delete"
    STATUS_UPDATE = "status_update"
    TASK_COMPLETE = "task_complete"
    TASK_CREATE = "task_create"
    TASK_DELETE = "task_delete"
    TASK_DOC_RELATIONSHIP_CREATE = "task_doc_relationship_create"
    TASK_DOC_RELATIONSHIP_DELETE = "task_doc_relationship_delete"
    TASK_RENAME = "task_rename"
    TASK_UPDATE = "task_update"
    TENANT_UPDATE = "tenant_update"
    USER_DARTBOARD_LAYOUT_CREATE = "user_dartboard_layout_create"
    USER_DARTBOARD_LAYOUT_DELETE = "user_dartboard_layout_delete"
    USER_DARTBOARD_LAYOUT_UPDATE = "user_dartboard_layout_update"
    USER_UPDATE = "user_update"
    VIEW_CREATE = "view_create"
    VIEW_DELETE = "view_delete"
    VIEW_UPDATE = "view_update"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class PropertyKind(str, Enum):
    CHECKBOX = "Checkbox"
    DATES = "Dates"
    DEFAULT_ASSIGNEE = "Default: Assignee"
    DEFAULT_CREATED = "Default: Created"
    DEFAULT_CREATED_BY = "Default: Created by"
    DEFAULT_DATES = "Default: Dates"
    DEFAULT_LAST_UPDATED = "Default: Last updated"
    DEFAULT_LAST_UPDATED_BY = "Default: Last updated by"
    DEFAULT_PRIORITY = "Default: Priority"
    DEFAULT_SIZE = "Default: Size"
    DEFAULT_STATUS = "Default: Status"
    DEFAULT_TAGS = "Default: Tags"
    MULTISELECT = "Multiselect"
    NUMBER = "Number"
    SELECT = "Select"
    STATUS = "Status"
    TEXT = "Text"
    USER = "User"

    def __str__(self) -> str:
        return str(self.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""A CLI to interact with the Dart web app."""

import argparse
from functools import wraps
import json
import os
import random
import re
import signal
import string
import subprocess
import sys
from collections import defaultdict
from datetime import timezone
from getpass import getpass
from importlib.metadata import version

import dateparser
from pick import pick
import requests

from .generated import Client
from .generated.models import (
    DartboardKind,
    Operation,
    OperationKind,
    OperationModelKind,
    Priority,
    PropertyKind,
    RequestBody,
    SpaceKind,
    StatusKind,
    TaskCreate,
    TaskSourceType,
    TaskUpdate,
    Transaction,
    TransactionKind,
)
from .generated.api.transactions import transactions_create
from .order_manager import get_orders_between

_PROG = "dart"

_PROD_HOST = "https://app.itsdart.com"
_DEV_HOST = "http://localhost:5173"
_HOST_MAP = {"prod": _PROD_HOST, "dev": _DEV_HOST}

_VERSION_CMD = "--version"
_SET_HOST_CMD = "sethost"
_LOGIN_CMD = "login"
_CREATE_TASK_CMD = "createtask"
_UPDATE_TASK_CMD = "updatetask"
_BEGIN_TASK_CMD = "begintask"

_ROOT_API_URL_FRAG = "/api/v0"
_CSRF_URL_FRAG = _ROOT_API_URL_FRAG + "/csrf-token"
_LOGIN_URL_FRAG = _ROOT_API_URL_FRAG + "/login"
_CURRENT_USER_URL_FRAG = _ROOT_API_URL_FRAG + "/current-user"
_COPY_BRANCH_URL_FRAG = _ROOT_API_URL_FRAG + "/vcs/copy-branch-link"

_CONFIG_FPATH = os.path.expanduser("~/.dart-tools")
_CSRF_TOKEN_COOKIE = "csrftoken"
_SESSION_ID_COOKIE = "sessionid"
_CLIENT_DUID_KEY = "clientDuid"
_HOST_KEY = "host"
_HOSTS_KEY = "hosts"
_CSRF_TOKEN_KEY = "csrfToken"
_SESSION_ID_KEY = "sessionId"

_DUID_CHARS = string.ascii_lowercase + string.ascii_uppercase + string.digits + "-_"
_PRIORITY_MAP = {
    0: Priority.CRITICAL,
    1: Priority.HIGH,
    2: Priority.MEDIUM,
    3: Priority.LOW,
}
_SIZES = {1, 2, 3, 5, 8}
_COMPLETED_STATUS_KINDS = {"Finished", "Canceled"}

_VERSION = version("dart-tools")

_is_cli = False


# TODO dedupe these functions with other usages elsewhere
def _make_duid():
    return "".join(random.choices(_DUID_CHARS, k=12))


def _run_cmd(cmd):
    return subprocess.check_output(cmd, shell=True).decode()


def _get_task_url(host, duid):
    return f"{host}/t/{duid}"


def _suppress_exception(fn):
    @wraps(fn)
    def wrapper(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:  # pylint: disable=broad-except
            return None

    return wrapper


def _exit_gracefully(_signal_received, _frame) -> None:
    sys.exit("Quitting.")


def _log(s):
    if not _is_cli:
        return
    print(s)


class _Config:
    def __init__(self):
        self._content = {}
        if os.path.isfile(_CONFIG_FPATH):
            with open(_CONFIG_FPATH, "r", encoding="UTF-8") as fin:
                self._content = json.load(fin)
        self._content = {
            _CLIENT_DUID_KEY: _make_duid(),
            _HOST_KEY: _PROD_HOST,
            _HOSTS_KEY: {},
        } | self._content
        self._content[_HOSTS_KEY] = defaultdict(dict, self._content[_HOSTS_KEY])
        self._write()

    def _write(self):
        with open(_CONFIG_FPATH, "w+", encoding="UTF-8") as fout:
            json.dump(self._content, fout, indent=2)

    @property
    def client_duid(self):
        return self._content[_CLIENT_DUID_KEY]

    @property
    def host(self):
        return self._content[_HOST_KEY]

    @host.setter
    def host(self, v):
        self._content[_HOST_KEY] = v
        self._write()

    def get(self, k):
        return self._content[_HOSTS_KEY][self.host].get(k)

    def set(self, k, v):
        self._content[_HOSTS_KEY][self.host][k] = v
        self._write()


class _Session:
    def __init__(self, config=None):
        self._config = config or _Config()
        self._session = requests.Session()
        if (csrf_token := self._config.get(_CSRF_TOKEN_KEY)) is not None:
            self._session.cookies.set(_CSRF_TOKEN_COOKIE, csrf_token)
        if (session_id := self._config.get(_SESSION_ID_KEY)) is not None:
            self._session.cookies.set(_SESSION_ID_COOKIE, session_id)

    def get_base_url(self):
        return self._config.host

    def get_client_duid(self):
        return self._config.client_duid

    def get_is_logged_in(self):
        return self._config.get(_SESSION_ID_KEY) is not None

    def _refresh_csrf(self):
        response = self._session.get(self._config.host + _CSRF_URL_FRAG)
        response.raise_for_status()
        csrf_token = response.cookies.get_dict().get(_CSRF_TOKEN_COOKIE)
        self._config.set(_CSRF_TOKEN_KEY, csrf_token)
        return csrf_token

    def get_csrf_token(self):
        csrf_token = self._config.get(_CSRF_TOKEN_KEY)
        if csrf_token is None:
            csrf_token = self._refresh_csrf()
        return csrf_token

    def get_headers(self):
        return {
            "client-duid": self._config.client_duid,
            "Origin": self._config.host,
            "x-csrftoken": self.get_csrf_token(),
        }

    def get_cookies(self):
        return {
            _SESSION_ID_COOKIE: self._config.get(_SESSION_ID_KEY),
            _CSRF_TOKEN_COOKIE: self.get_csrf_token(),
        }

    def get(self, url_frag, *args, **kwargs):
        kwargs["headers"] = self.get_headers() | kwargs.get("headers", {})
        return self._session.get(self._config.host + url_frag, *args, **kwargs)

    def post(self, url_frag, *args, **kwargs):
        kwargs["headers"] = self.get_headers() | kwargs.get("headers", {})
        result = self._session.post(self._config.host + url_frag, *args, **kwargs)
        if result.status_code != 403:
            return result
        self._refresh_csrf()
        kwargs["headers"] = self.get_headers() | kwargs.get("headers", {})
        return self._session.post(self._config.host + url_frag, *args, **kwargs)


class Dart(Client):
    def __init__(self, session=None):
        self._session = session or _Session()
        super().__init__(
            base_url=self._session.get_base_url(),
            cookies=self._session.get_cookies(),
            headers=self._session.get_headers(),
        )

    def transact(self, operations: list[Operation], kind: TransactionKind):
        transaction = Transaction(
            duid=_make_duid(),
            kind=kind,
            operations=operations,
        )
        request_body = RequestBody(
            client_duid=self._session.get_client_duid(),
            items=[transaction],
        )
        return transactions_create.sync(
            client=self,
            x_csrftoken=self._session.get_csrf_token(),
            json_body=request_body,
        )


class UserBundle:
    def __init__(self, session):
        _log("Loading active tasks")
        response = session.get(_CURRENT_USER_URL_FRAG)
        _check_request_response_and_maybe_exit(response)
        self._raw = response.json()
        if not self.is_logged_in:
            _auth_failure_exit()

    @property
    def is_logged_in(self):
        return self._raw["isLoggedIn"]

    @property
    def user(self):
        return self._raw["user"]

    @property
    def users(self):
        return self._raw["users"]

    @property
    def properties(self):
        return self._raw["properties"]

    @property
    def default_statuses(self):
        default_status_property_duid = next(
            e["duid"]
            for e in self.properties
            if e["kind"] == PropertyKind.DEFAULT_STATUS
        )
        return [
            e
            for e in self._raw["statuses"]
            if e["propertyDuid"] == default_status_property_duid
        ]

    @property
    def default_tags(self):
        default_tags_property_duid = next(
            e["duid"] for e in self.properties if e["kind"] == PropertyKind.DEFAULT_TAGS
        )
        return [
            e
            for e in self._raw["options"]
            if e["propertyDuid"] == default_tags_property_duid
        ]

    @property
    def spaces(self):
        return self._raw["spaces"]

    @property
    def dartboards(self):
        return self._raw["dartboards"]

    @property
    def tasks(self):
        return self._raw["tasks"]


class _Git:
    @staticmethod
    def _cmd_succeeds(cmd):
        try:
            _run_cmd(f"{cmd} 2>&1")
        except subprocess.CalledProcessError as ex:
            if "128" in str(ex):
                return False
            raise ex
        return True

    @staticmethod
    def _format_for_branch(s):
        return re.sub(
            r"-{2,}", "-", re.sub(r"[^a-z0-9-]+", "-", s.lower().replace("'", ""))
        ).strip("-")

    @staticmethod
    def make_task_name(email, task):
        username = _Git._format_for_branch(email.split("@")[0])
        title = _Git._format_for_branch(task.title)
        long = f"{username}/{task.duid}-{title}"
        if len(long) <= 60:
            return long
        for i in range(1, 11):
            if long[60 - i] == "-":
                return long[: 60 - i]
        return long[:60]

    @staticmethod
    def get_current_branch():
        return _run_cmd("git rev-parse --abbrev-ref HEAD").strip()

    @staticmethod
    def ensure_in_repo():
        if _Git._cmd_succeeds("git rev-parse --is-inside-work-tree"):
            return
        sys.exit("You are not in a git repo.")

    @staticmethod
    def ensure_no_unstaged_changes():
        if _run_cmd("git status --porcelain") == "":
            return
        sys.exit("You have uncommitted changes. Please commit or stash them.")

    @staticmethod
    def ensure_on_main_or_intended():
        branch = _Git.get_current_branch()
        if branch == "main":
            return
        if (
            pick(
                ["Yes", "No"],
                "You're not on the 'main' branch. Is this intentional?",
                "→",
            )[0]
            == "Yes"
        ):
            return
        _run_cmd("git checkout main")

    @staticmethod
    def branch_exists(branch):
        return _Git._cmd_succeeds(f"git rev-parse --verify {branch}")

    @staticmethod
    def checkout_branch(branch):
        if _Git.branch_exists(branch):
            _run_cmd(f"git checkout {branch}")
            return

        if _Git.branch_exists(f"origin/{branch}"):
            _run_cmd(f"git checkout --track origin/{branch}")
            return

        _run_cmd(f"git checkout -b {branch}")


def set_host(host):
    config = _Config()

    new_host = _HOST_MAP.get(host, host)
    _log(f"Setting host to {new_host}")
    config.host = new_host

    _log("Done.")
    return True


def _auth_failure_exit():
    sys.exit(f"Not logged in, run\n\n{_PROG} {_LOGIN_CMD}\n\nto log in.")


def _unknown_failure_exit():
    sys.exit(f"Not logged in, run\n\n{_PROG} {_LOGIN_CMD}\n\nto log in.")


def _check_request_response_and_maybe_exit(response):
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError:
        if response.status_code in {401, 403}:
            _auth_failure_exit()
        _unknown_failure_exit()


def _parse_transaction_response_and_maybe_exit(response, model_kind, duid):
    if (
        response is None
        or not hasattr(response, "results")
        or len(response.results) == 0
        or not response.results[0].success
    ):
        _unknown_failure_exit()
    models = getattr(response.results[0].models, f"{model_kind}s")
    model = next((e for e in models if e.duid == duid), None)
    if model is None:
        _unknown_failure_exit()
    return model


def print_version():
    result = f"dart-tools version {_VERSION}"
    _log(result)
    return result


@_suppress_exception
def print_version_update_message_maybe():
    latest = (
        _run_cmd("pip --disable-pip-version-check index versions dart-tools 2>&1")
        .rsplit("LATEST:", maxsplit=1)[-1]
        .split("\n", maxsplit=1)[0]
        .strip()
    )
    if latest == _VERSION or [int(e) for e in latest.split(".")] <= [
        int(e) for e in _VERSION.split(".")
    ]:
        return

    _log(
        f"A new version of dart-tools is available. Upgrade from {_VERSION} to {latest} with\n\n  pip install --upgrade dart-tools\n"
    )


def is_logged_in():
    config = _Config()
    session = _Session(config)

    result = session.get_is_logged_in()
    _log(f"You are {'' if result else 'not '}logged in")
    return result


def login(*, email=None, password=None):
    config = _Config()
    session = _Session(config)

    _log("Log in to Dart")
    if email is None:
        email = input("Email: ")
    if password is None:
        password = getpass()

    result = session.post(_LOGIN_URL_FRAG, json={"email": email, "password": password})
    if result.status_code in {401, 403}:
        sys.exit("Invalid login information.")
    _check_request_response_and_maybe_exit(result)

    cookies = result.cookies.get_dict()
    config.set(_SESSION_ID_KEY, cookies.get(_SESSION_ID_COOKIE))

    _log("Logged in.")
    return True


def _begin_task(config, session, user_email, get_task):
    _Git.ensure_in_repo()
    _Git.ensure_no_unstaged_changes()
    _Git.ensure_on_main_or_intended()

    task = get_task()

    response = session.post(_COPY_BRANCH_URL_FRAG, json={"duid": task.duid})
    _check_request_response_and_maybe_exit(response)

    branch_name = _Git.make_task_name(user_email, task)
    _Git.checkout_branch(branch_name)

    _log(
        f"Started work on\n\n  {task.title}\n  {_get_task_url(config.host, task.duid)}\n"
    )
    return True


def begin_task():
    config = _Config()
    session = _Session(config)

    user_bundle = UserBundle(session)
    user = user_bundle.user

    def _get_task():
        user_duid = user["duid"]
        team_space_duid = next(
            e["duid"] for e in user_bundle.spaces if e["kind"] == SpaceKind.WORKSPACE
        )
        active_duid = next(
            e["duid"]
            for e in user_bundle.dartboards
            if e["spaceDuid"] == team_space_duid and e["kind"] == DartboardKind.ACTIVE
        )
        unterm_status_duids = {
            e["duid"]
            for e in user_bundle.default_statuses
            if e["kind"] not in _COMPLETED_STATUS_KINDS
        }
        filtered_tasks = [
            e
            for e in user_bundle.tasks
            if not e["inTrash"]
            and e["dartboardDuid"] == active_duid
            and user_duid in e["assigneeDuids"]
            and e["statusDuid"] in unterm_status_duids
            and e["drafterDuid"] is None
        ]
        filtered_tasks.sort(key=lambda e: e["order"])

        if len(filtered_tasks) == 0:
            sys.exit("No active, incomplete tasks found.")

        return TaskCreate.from_dict(
            filtered_tasks[
                pick(
                    [e["title"] for e in filtered_tasks],
                    "Which of your active, incomplete tasks are you beginning work on?",
                    "→",
                )[1]
            ]
        )

    _begin_task(config, session, user["email"], _get_task)

    _log("Done.")
    return True


def create_task(
    title,
    *,
    should_begin=False,
    dartboard_title=None,
    status_title=None,
    assignee_emails=None,
    tag_titles=None,
    priority_int=None,
    size_int=None,
    due_at_str=None,
):
    config = _Config()
    session = _Session(config)
    dart = Dart(session)

    user_bundle = UserBundle(session)

    user = user_bundle.user
    user_duid = user["duid"]

    dartboards = user_bundle.dartboards
    if dartboard_title is not None:
        dartboard_title_norm = dartboard_title.strip().lower()
        dartboard = next(
            (
                e
                for e in dartboards
                if dartboard_title_norm in {e["title"].lower(), e["kind"].lower()}
            ),
            None,
        )
        if dartboard is None:
            sys.exit(f"No dartboard found with title '{dartboard_title}'.")
    else:
        dartboard = next(e for e in dartboards if e["kind"] == DartboardKind.ACTIVE)
    dartboard_duid = dartboard["duid"]

    orders = [
        e["order"] for e in user_bundle.tasks if e["dartboardDuid"] == dartboard_duid
    ]
    first_order = min(orders) if len(orders) > 0 else None
    order = get_orders_between(None, first_order, 1)[0]

    statuses = user_bundle.default_statuses
    if status_title is not None:
        status_title_norm = status_title.strip().lower()
        status = next(
            (e for e in statuses if e["title"].lower() == status_title_norm), None
        )
        if status is None:
            sys.exit(f"No status found with title '{status_title}'.")
    else:
        status = next(
            e for e in statuses if e["kind"] == StatusKind.UNSTARTED and e["locked"]
        )
    status_duid = status["duid"]

    users = user_bundle.users
    user_emails_to_duids = {e["email"]: e["duid"] for e in users}
    assignee_duids = []
    subscriber_duids = []
    if assignee_emails is not None:
        for assignee_email in assignee_emails:
            assignee_email_norm = assignee_email.strip().lower()
            if assignee_email_norm not in user_emails_to_duids:
                sys.exit(f"No user found with email '{assignee_email}'.")
            assignee_duids.append(user_emails_to_duids[assignee_email_norm])
            subscriber_duids.append(user_emails_to_duids[assignee_email_norm])
    else:
        assignee_duids.append(user_duid)
    assignee_duids = list(set(assignee_duids))
    subscriber_duids.append(user_duid)
    subscriber_duids = list(set(subscriber_duids))

    tags = user_bundle.default_tags
    tag_titles_to_duids = {e["title"]: e["duid"] for e in tags}
    tag_duids = []
    if tag_titles is not None:
        for tag_title in tag_titles:
            tag_title_norm = tag_title.strip().lower()
            if tag_title_norm not in tag_titles_to_duids:
                sys.exit(f"No tag found with title '{tag_title}'.")
            tag_duids.append(tag_titles_to_duids[tag_title_norm])

    priority = None
    if priority_int is not None:
        priority = _PRIORITY_MAP[priority_int]

    size = size_int

    due_at = None
    if due_at_str is not None:
        due_at = dateparser.parse(due_at_str)
        if due_at is None:
            sys.exit(f"Could not parse due date '{due_at_str}'.")
        due_at = due_at.replace(
            hour=9, minute=0, second=0, microsecond=0, tzinfo=timezone.utc
        )

    task_create = TaskCreate(
        duid=_make_duid(),
        source_type=TaskSourceType.CLI,
        drafter_duid=None,
        dartboard_duid=dartboard_duid,
        order=order,
        title=title,
        status_duid=status_duid,
        assignee_duids=assignee_duids,
        subscriber_duids=subscriber_duids,
        tag_duids=tag_duids,
        priority=priority,
        size=size,
        due_at=due_at,
    )
    task_create_op = Operation(
        model=OperationModelKind.TASK,
        kind=OperationKind.CREATE,
        data=task_create,
    )
    response = dart.transact([task_create_op], TransactionKind.TASK_CREATE)
    task = _parse_transaction_response_and_maybe_exit(
        response, OperationModelKind.TASK, task_create.duid
    )

    _log(f"Created task {task.title} at {_get_task_url(config.host, task.duid)}")

    if should_begin:
        _begin_task(config, session, user["email"], lambda: task_create)

    _log("Done.")
    return task


def update_task(
    duid,
    *,
    title=None,
    dartboard_title=None,
    status_title=None,
    assignee_emails=None,
    tag_titles=None,
    priority_int=None,
    size_int=None,
    due_at_str=None,
):
    config = _Config()
    session = _Session(config)
    dart = Dart(session)

    user_bundle = UserBundle(session)

    user = user_bundle.user
    user_duid = user["duid"]

    tasks = user_bundle.tasks
    old_task = next((e for e in tasks if e["duid"] == duid), None)
    if old_task is None:
        sys.exit(f"No task found with DUID '{duid}'.")

    task_update_kwargs = {"duid": duid}

    if title is not None:
        task_update_kwargs["title"] = title

    dartboards = user_bundle.dartboards
    if dartboard_title is not None:
        dartboard_title_norm = dartboard_title.strip().lower()
        dartboard = next(
            (
                e
                for e in dartboards
                if dartboard_title_norm in {e["title"].lower(), e["kind"].lower()}
            ),
            None,
        )
        if dartboard is None:
            sys.exit(f"No dartboard found with title '{dartboard_title}'.")
        dartboard_duid = dartboard["duid"]
        if dartboard_duid != old_task["dartboardDuid"]:
            task_update_kwargs["dartboard_duid"] = dartboard_duid

    statuses = user_bundle.default_statuses
    if status_title is not None:
        status_title_norm = status_title.strip().lower()
        status = next(
            (e for e in statuses if e["title"].lower() == status_title_norm), None
        )
        if status is None:
            sys.exit(f"No status found with title '{status_title}'.")
        status_duid = status["duid"]
        if status_duid != old_task["statusDuid"]:
            task_update_kwargs["status_duid"] = status_duid

    users = user_bundle.users
    user_emails_to_duids = {e["email"]: e["duid"] for e in users}
    subscriber_duids = []
    if assignee_emails is not None:
        assignee_duids = []
        for assignee_email in assignee_emails:
            assignee_email_norm = assignee_email.strip().lower()
            if assignee_email_norm not in user_emails_to_duids:
                sys.exit(f"No user found with email '{assignee_email}'.")
            assignee_duids.append(user_emails_to_duids[assignee_email_norm])
            subscriber_duids.append(user_emails_to_duids[assignee_email_norm])
        assignee_duids = sorted(set(assignee_duids))
        if assignee_duids != old_task["assigneeDuids"]:
            task_update_kwargs["assignee_duids"] = assignee_duids

    # TODO do add to list operation rather than replace
    subscriber_duids = list(
        set(old_task["subscriberDuids"]) | set(subscriber_duids) | {user_duid}
    )
    if subscriber_duids != old_task["subscriberDuids"]:
        task_update_kwargs["subscriber_duids"] = subscriber_duids

    tags = user_bundle.default_tags
    tag_titles_to_duids = {e["title"]: e["duid"] for e in tags}
    if tag_titles is not None:
        tag_duids = []
        for tag_title in tag_titles:
            tag_title_norm = tag_title.strip().lower()
            if tag_title_norm not in tag_titles_to_duids:
                sys.exit(f"No tag found with title '{tag_title}'.")
            tag_duids.append(tag_titles_to_duids[tag_title_norm])
        task_update_kwargs["tag_duids"] = tag_duids

    # TODO add a way for optional stuff to be removed
    if priority_int is not None:
        priority = _PRIORITY_MAP[priority_int]
        if priority != old_task["priority"]:
            task_update_kwargs["priority"] = priority

    if size_int is not None:
        size = size_int
        if size != old_task["size"]:
            task_update_kwargs["size"] = size

    if due_at_str is not None:
        due_at = dateparser.parse(due_at_str)
        if due_at is None:
            sys.exit(f"Could not parse due date '{due_at_str}'.")
        due_at = due_at.replace(
            hour=9, minute=0, second=0, microsecond=0, tzinfo=timezone.utc
        )
        due_at = due_at.isoformat()[:-6] + ".000Z"
        if due_at != old_task["dueAt"]:
            task_update_kwargs["due_at"] = due_at

    task_update = TaskUpdate(**task_update_kwargs)
    task_update_op = Operation(
        model=OperationModelKind.TASK,
        kind=OperationKind.UPDATE,
        data=task_update,
    )
    response = dart.transact([task_update_op], TransactionKind.TASK_UPDATE)
    task = _parse_transaction_response_and_maybe_exit(
        response, OperationModelKind.TASK, task_update.duid
    )

    _log(f"Updated task {task.title} at {_get_task_url(config.host, task.duid)}")
    _log("Done.")
    return task


def _add_standard_task_arguments(parser):
    parser.add_argument(
        "-d", "--dartboard", dest="dartboard_title", help="dartboard title"
    )
    parser.add_argument("-s", "--status", dest="status_title", help="status title")
    parser.add_argument(
        "-a",
        "--assignee",
        dest="assignee_emails",
        nargs="*",
        action="extend",
        help="assignee email(s)",
    )
    parser.add_argument(
        "-t",
        "--tag",
        dest="tag_titles",
        nargs="*",
        action="extend",
        help="tag title(s)",
    )
    parser.add_argument(
        "-p",
        "--priority",
        dest="priority_int",
        type=int,
        choices=_PRIORITY_MAP.keys(),
        help="priority",
    )
    parser.add_argument(
        "-i", "--size", dest="size_int", type=int, choices=_SIZES, help="size"
    )
    parser.add_argument("-r", "--duedate", dest="due_at_str", help="due date")


def cli():
    signal.signal(signal.SIGINT, _exit_gracefully)
    global _is_cli
    _is_cli = True

    print_version_update_message_maybe()

    if _VERSION_CMD in sys.argv[1:]:
        print_version()
        return

    parser = argparse.ArgumentParser(
        prog=_PROG, description="A CLI to interact with Dart"
    )
    subparsers = parser.add_subparsers(
        title="command",
        required=True,
        metavar=f"{{{_LOGIN_CMD},{_CREATE_TASK_CMD},{_BEGIN_TASK_CMD}}}",
    )

    set_host_parser = subparsers.add_parser(_SET_HOST_CMD, aliases="s")
    set_host_parser.add_argument("host", help="the new host: {prod|dev|[URL]}")
    set_host_parser.set_defaults(func=set_host)

    login_parser = subparsers.add_parser(_LOGIN_CMD, aliases="l", help="login")
    login_parser.add_argument(
        "-e", "--email", dest="email", help="email to log in with"
    )
    login_parser.add_argument(
        "-p", "--password", dest="password", help="password to log in with"
    )
    login_parser.set_defaults(func=login)

    create_task_parser = subparsers.add_parser(
        _CREATE_TASK_CMD, aliases="c", help="create a new task"
    )
    create_task_parser.add_argument("title", help="title of the task")
    create_task_parser.add_argument(
        "-b",
        "--begin",
        dest="should_begin",
        action="store_true",
        help="begin work on the task after creation",
    )
    _add_standard_task_arguments(create_task_parser)
    create_task_parser.set_defaults(func=create_task)

    update_task_parser = subparsers.add_parser(
        _UPDATE_TASK_CMD, aliases="u", help="update an existing task"
    )
    update_task_parser.add_argument("duid", help="Dart ID (DUID) of the task")
    update_task_parser.add_argument("-e", "--title", dest="title", help="task title")
    _add_standard_task_arguments(update_task_parser)
    update_task_parser.set_defaults(func=update_task)

    begin_task_parser = subparsers.add_parser(
        _BEGIN_TASK_CMD, aliases="b", help="begin work on a task"
    )
    begin_task_parser.set_defaults(func=begin_task)

    args = vars(parser.parse_args())
    func = args.pop("func")
    func(**args)



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.priority import Priority
from ..models.recommendation_status import RecommendationStatus
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.relationship import Relationship
    from ..models.task_attachment import TaskAttachment
    from ..models.task_description import TaskDescription
    from ..models.task_link import TaskLink
    from ..models.task_notion_document import TaskNotionDocument
    from ..models.task_properties import TaskProperties
    from ..models.task_recurrence import TaskRecurrence


T = TypeVar("T", bound="Task")


@_attrs_define
class Task:
    """
    Attributes:
        duid (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        in_trash (bool):
        dartboard_duid (str):
        order (str):
        title (str):
        description (TaskDescription):
        status_duid (str):
        assigned_to_ai (bool):
        assignee_duids (List[str]):
        subscriber_duids (List[str]):
        tag_duids (List[str]):
        links (List['TaskLink']):
        attachments (List['TaskAttachment']):
        relationships (List['Relationship']):
        properties (TaskProperties):
        updated_by_client_duid (Union[Unset, None, str]):
        drafter_duid (Optional[str]):
        recommendation_status (Optional[RecommendationStatus]): * `Accepted` - ACCEPTED
            * `Declined` - DECLINED
        priority (Optional[Priority]): * `Critical` - CRITICAL
            * `High` - HIGH
            * `Medium` - MEDIUM
            * `Low` - LOW
        size (Optional[int]):
        start_at (Optional[datetime.datetime]):
        due_at (Optional[datetime.datetime]):
        remind_at (Optional[datetime.datetime]):
        recurrence (Optional[TaskRecurrence]):
        recurrs_next_at (Optional[datetime.datetime]):
        notion_document (Optional[TaskNotionDocument]):
    """

    duid: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    in_trash: bool
    dartboard_duid: str
    order: str
    title: str
    description: "TaskDescription"
    status_duid: str
    assigned_to_ai: bool
    assignee_duids: List[str]
    subscriber_duids: List[str]
    tag_duids: List[str]
    links: List["TaskLink"]
    attachments: List["TaskAttachment"]
    relationships: List["Relationship"]
    properties: "TaskProperties"
    drafter_duid: Optional[str]
    recommendation_status: Optional[RecommendationStatus]
    priority: Optional[Priority]
    size: Optional[int]
    start_at: Optional[datetime.datetime]
    due_at: Optional[datetime.datetime]
    remind_at: Optional[datetime.datetime]
    recurrence: Optional["TaskRecurrence"]
    recurrs_next_at: Optional[datetime.datetime]
    notion_document: Optional["TaskNotionDocument"]
    updated_by_client_duid: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        in_trash = self.in_trash
        dartboard_duid = self.dartboard_duid
        order = self.order
        title = self.title
        description = self.description.to_dict()

        status_duid = self.status_duid
        assigned_to_ai = self.assigned_to_ai
        assignee_duids = self.assignee_duids

        subscriber_duids = self.subscriber_duids

        tag_duids = self.tag_duids

        links = []
        for links_item_data in self.links:
            links_item = links_item_data.to_dict()

            links.append(links_item)

        attachments = []
        for attachments_item_data in self.attachments:
            attachments_item = attachments_item_data.to_dict()

            attachments.append(attachments_item)

        relationships = []
        for relationships_item_data in self.relationships:
            relationships_item = relationships_item_data.to_dict()

            relationships.append(relationships_item)

        properties = self.properties.to_dict()

        updated_by_client_duid = self.updated_by_client_duid
        drafter_duid = self.drafter_duid
        recommendation_status = self.recommendation_status.value if self.recommendation_status else None

        priority = self.priority.value if self.priority else None

        size = self.size
        start_at = self.start_at.isoformat() if self.start_at else None

        due_at = self.due_at.isoformat() if self.due_at else None

        remind_at = self.remind_at.isoformat() if self.remind_at else None

        recurrence = self.recurrence.to_dict() if self.recurrence else None

        recurrs_next_at = self.recurrs_next_at.isoformat() if self.recurrs_next_at else None

        notion_document = self.notion_document.to_dict() if self.notion_document else None

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "createdAt": created_at,
                "updatedAt": updated_at,
                "inTrash": in_trash,
                "dartboardDuid": dartboard_duid,
                "order": order,
                "title": title,
                "description": description,
                "statusDuid": status_duid,
                "assignedToAi": assigned_to_ai,
                "assigneeDuids": assignee_duids,
                "subscriberDuids": subscriber_duids,
                "tagDuids": tag_duids,
                "links": links,
                "attachments": attachments,
                "relationships": relationships,
                "properties": properties,
                "drafterDuid": drafter_duid,
                "recommendationStatus": recommendation_status,
                "priority": priority,
                "size": size,
                "startAt": start_at,
                "dueAt": due_at,
                "remindAt": remind_at,
                "recurrence": recurrence,
                "recurrsNextAt": recurrs_next_at,
                "notionDocument": notion_document,
            }
        )
        if updated_by_client_duid is not UNSET:
            field_dict["updatedByClientDuid"] = updated_by_client_duid

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.relationship import Relationship
        from ..models.task_attachment import TaskAttachment
        from ..models.task_description import TaskDescription
        from ..models.task_link import TaskLink
        from ..models.task_notion_document import TaskNotionDocument
        from ..models.task_properties import TaskProperties
        from ..models.task_recurrence import TaskRecurrence

        d = src_dict.copy()
        duid = d.pop("duid")

        created_at = isoparse(d.pop("createdAt"))

        updated_at = isoparse(d.pop("updatedAt"))

        in_trash = d.pop("inTrash")

        dartboard_duid = d.pop("dartboardDuid")

        order = d.pop("order")

        title = d.pop("title")

        description = TaskDescription.from_dict(d.pop("description"))

        status_duid = d.pop("statusDuid")

        assigned_to_ai = d.pop("assignedToAi")

        assignee_duids = cast(List[str], d.pop("assigneeDuids"))

        subscriber_duids = cast(List[str], d.pop("subscriberDuids"))

        tag_duids = cast(List[str], d.pop("tagDuids"))

        links = []
        _links = d.pop("links")
        for links_item_data in _links:
            links_item = TaskLink.from_dict(links_item_data)

            links.append(links_item)

        attachments = []
        _attachments = d.pop("attachments")
        for attachments_item_data in _attachments:
            attachments_item = TaskAttachment.from_dict(attachments_item_data)

            attachments.append(attachments_item)

        relationships = []
        _relationships = d.pop("relationships")
        for relationships_item_data in _relationships:
            relationships_item = Relationship.from_dict(relationships_item_data)

            relationships.append(relationships_item)

        properties = TaskProperties.from_dict(d.pop("properties"))

        updated_by_client_duid = d.pop("updatedByClientDuid", UNSET)

        drafter_duid = d.pop("drafterDuid")

        _recommendation_status = d.pop("recommendationStatus")
        recommendation_status: Optional[RecommendationStatus]
        if _recommendation_status is None:
            recommendation_status = None
        else:
            recommendation_status = RecommendationStatus(_recommendation_status)

        _priority = d.pop("priority")
        priority: Optional[Priority]
        if _priority is None:
            priority = None
        else:
            priority = Priority(_priority)

        size = d.pop("size")

        _start_at = d.pop("startAt")
        start_at: Optional[datetime.datetime]
        if _start_at is None:
            start_at = None
        else:
            start_at = isoparse(_start_at)

        _due_at = d.pop("dueAt")
        due_at: Optional[datetime.datetime]
        if _due_at is None:
            due_at = None
        else:
            due_at = isoparse(_due_at)

        _remind_at = d.pop("remindAt")
        remind_at: Optional[datetime.datetime]
        if _remind_at is None:
            remind_at = None
        else:
            remind_at = isoparse(_remind_at)

        _recurrence = d.pop("recurrence")
        recurrence: Optional[TaskRecurrence]
        if _recurrence is None:
            recurrence = None
        else:
            recurrence = TaskRecurrence.from_dict(_recurrence)

        _recurrs_next_at = d.pop("recurrsNextAt")
        recurrs_next_at: Optional[datetime.datetime]
        if _recurrs_next_at is None:
            recurrs_next_at = None
        else:
            recurrs_next_at = isoparse(_recurrs_next_at)

        _notion_document = d.pop("notionDocument")
        notion_document: Optional[TaskNotionDocument]
        if _notion_document is None:
            notion_document = None
        else:
            notion_document = TaskNotionDocument.from_dict(_notion_document)

        task = cls(
            duid=duid,
            created_at=created_at,
            updated_at=updated_at,
            in_trash=in_trash,
            dartboard_duid=dartboard_duid,
            order=order,
            title=title,
            description=description,
            status_duid=status_duid,
            assigned_to_ai=assigned_to_ai,
            assignee_duids=assignee_duids,
            subscriber_duids=subscriber_duids,
            tag_duids=tag_duids,
            links=links,
            attachments=attachments,
            relationships=relationships,
            properties=properties,
            updated_by_client_duid=updated_by_client_duid,
            drafter_duid=drafter_duid,
            recommendation_status=recommendation_status,
            priority=priority,
            size=size,
            start_at=start_at,
            due_at=due_at,
            remind_at=remind_at,
            recurrence=recurrence,
            recurrs_next_at=recurrs_next_at,
            notion_document=notion_document,
        )

        task.additional_properties = d
        return task

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Optional, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.github_integration_tenant_extension_status import GithubIntegrationTenantExtensionStatus

T = TypeVar("T", bound="GithubIntegration")


@_attrs_define
class GithubIntegration:
    """
    Attributes:
        status (GithubIntegrationTenantExtensionStatus): * `Disabled` - DISABLED
            * `Pending` - PENDING
            * `Suspended` - SUSPENDED
            * `Enabled` - ENABLED
        linkback_enabled (bool):
        installation_link (Optional[str]):
        status_on_branch_link_copy_duid (Optional[str]):
        status_on_branch_create_duid (Optional[str]):
        status_on_pr_draft_duid (Optional[str]):
        status_on_pr_ready_duid (Optional[str]):
        status_on_pr_merge_duid (Optional[str]):
    """

    status: GithubIntegrationTenantExtensionStatus
    linkback_enabled: bool
    installation_link: Optional[str]
    status_on_branch_link_copy_duid: Optional[str]
    status_on_branch_create_duid: Optional[str]
    status_on_pr_draft_duid: Optional[str]
    status_on_pr_ready_duid: Optional[str]
    status_on_pr_merge_duid: Optional[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        status = self.status.value

        linkback_enabled = self.linkback_enabled
        installation_link = self.installation_link
        status_on_branch_link_copy_duid = self.status_on_branch_link_copy_duid
        status_on_branch_create_duid = self.status_on_branch_create_duid
        status_on_pr_draft_duid = self.status_on_pr_draft_duid
        status_on_pr_ready_duid = self.status_on_pr_ready_duid
        status_on_pr_merge_duid = self.status_on_pr_merge_duid

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "status": status,
                "linkbackEnabled": linkback_enabled,
                "installationLink": installation_link,
                "statusOnBranchLinkCopyDuid": status_on_branch_link_copy_duid,
                "statusOnBranchCreateDuid": status_on_branch_create_duid,
                "statusOnPrDraftDuid": status_on_pr_draft_duid,
                "statusOnPrReadyDuid": status_on_pr_ready_duid,
                "statusOnPrMergeDuid": status_on_pr_merge_duid,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        status = GithubIntegrationTenantExtensionStatus(d.pop("status"))

        linkback_enabled = d.pop("linkbackEnabled")

        installation_link = d.pop("installationLink")

        status_on_branch_link_copy_duid = d.pop("statusOnBranchLinkCopyDuid")

        status_on_branch_create_duid = d.pop("statusOnBranchCreateDuid")

        status_on_pr_draft_duid = d.pop("statusOnPrDraftDuid")

        status_on_pr_ready_duid = d.pop("statusOnPrReadyDuid")

        status_on_pr_merge_duid = d.pop("statusOnPrMergeDuid")

        github_integration = cls(
            status=status,
            linkback_enabled=linkback_enabled,
            installation_link=installation_link,
            status_on_branch_link_copy_duid=status_on_branch_link_copy_duid,
            status_on_branch_create_duid=status_on_branch_create_duid,
            status_on_pr_draft_duid=status_on_pr_draft_duid,
            status_on_pr_ready_duid=status_on_pr_ready_duid,
            status_on_pr_merge_duid=status_on_pr_merge_duid,
        )

        github_integration.additional_properties = d
        return github_integration

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class LayoutKind(str, Enum):
    BOARD = "board"
    CALENDAR = "calendar"
    LIST = "list"
    ROADMAP = "roadmap"

    def __str__(self) -> str:
        return str(self.value)



from enum import Enum


class RelationshipKindKind(str, Enum):
    BLOCKS = "Blocks"
    CUSTOM = "Custom"
    DUPLICATES = "Duplicates"
    PARENT_OF = "Parent Of"
    RELATES_TO = "Relates To"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.transaction_kind import TransactionKind

if TYPE_CHECKING:
    from ..models.operation import Operation


T = TypeVar("T", bound="Transaction")


@_attrs_define
class Transaction:
    """
    Attributes:
        duid (str):
        kind (TransactionKind): * `comment_create` - COMMENT_CREATE
            * `comment_update` - COMMENT_UPDATE
            * `comment_delete` - COMMENT_DELETE
            * `comment_reaction_create` - COMMENT_REACTION_CREATE
            * `comment_reaction_delete` - COMMENT_REACTION_DELETE
            * `cycle_rollover` - CYCLE_ROLLOVER
            * `dartboard_create` - DARTBOARD_CREATE
            * `dartboard_delete` - DARTBOARD_DELETE
            * `dartboard_update` - DARTBOARD_UPDATE
            * `doc_create` - DOC_CREATE
            * `doc_delete` - DOC_DELETE
            * `doc_update` - DOC_UPDATE
            * `event_create` - EVENT_CREATE
            * `folder_create` - FOLDER_CREATE
            * `folder_delete` - FOLDER_DELETE
            * `folder_update` - FOLDER_UPDATE
            * `layout_create` - LAYOUT_CREATE
            * `layout_delete` - LAYOUT_DELETE
            * `layout_update` - LAYOUT_UPDATE
            * `option_create` - OPTION_CREATE
            * `option_delete` - OPTION_DELETE
            * `option_update` - OPTION_UPDATE
            * `property_create` - PROPERTY_CREATE
            * `property_update` - PROPERTY_UPDATE
            * `property_delete` - PROPERTY_DELETE
            * `relationship_create` - RELATIONSHIP_CREATE
            * `relationship_update` - RELATIONSHIP_UPDATE
            * `relationship_delete` - RELATIONSHIP_DELETE
            * `space_create` - SPACE_CREATE
            * `space_update_perms` - SPACE_UPDATE_PERMS
            * `space_update_other` - SPACE_UPDATE_OTHER
            * `space_delete` - SPACE_DELETE
            * `status_create` - STATUS_CREATE
            * `status_delete` - STATUS_DELETE
            * `status_update` - STATUS_UPDATE
            * `task_complete` - TASK_COMPLETE
            * `task_create` - TASK_CREATE
            * `task_delete` - TASK_DELETE
            * `task_rename` - TASK_RENAME
            * `task_update` - TASK_UPDATE
            * `task_doc_relationship_create` - TASK_DOC_RELATIONSHIP_CREATE
            * `task_doc_relationship_delete` - TASK_DOC_RELATIONSHIP_DELETE
            * `tenant_update` - TENANT_UPDATE
            * `user_dartboard_layout_create` - USER_DARTBOARD_LAYOUT_CREATE
            * `user_dartboard_layout_delete` - USER_DARTBOARD_LAYOUT_DELETE
            * `user_dartboard_layout_update` - USER_DARTBOARD_LAYOUT_UPDATE
            * `user_update` - USER_UPDATE
            * `view_create` - VIEW_CREATE
            * `view_delete` - VIEW_DELETE
            * `view_update` - VIEW_UPDATE
        operations (List['Operation']):
    """

    duid: str
    kind: TransactionKind
    operations: List["Operation"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        kind = self.kind.value

        operations = []
        for operations_item_data in self.operations:
            operations_item = operations_item_data.to_dict()

            operations.append(operations_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "kind": kind,
                "operations": operations,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.operation import Operation

        d = src_dict.copy()
        duid = d.pop("duid")

        kind = TransactionKind(d.pop("kind"))

        operations = []
        _operations = d.pop("operations")
        for operations_item_data in _operations:
            operations_item = Operation.from_dict(operations_item_data)

            operations.append(operations_item)

        transaction = cls(
            duid=duid,
            kind=kind,
            operations=operations,
        )

        transaction.additional_properties = d
        return transaction

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.color_name import ColorName
from ..models.icon_kind import IconKind
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.doc_text import DocText


T = TypeVar("T", bound="Doc")


@_attrs_define
class Doc:
    """
    Attributes:
        duid (str):
        created_at (datetime.datetime):
        updated_at (datetime.datetime):
        in_trash (bool):
        folder_duid (str):
        order (str):
        title (str):
        text (DocText):
        edited_by_ai (bool):
        editor_duids (List[str]):
        subscriber_duids (List[str]):
        icon_kind (IconKind): * `None` - NONE
            * `Icon` - ICON
            * `Emoji` - EMOJI
        icon_name_or_emoji (str):
        color_name (ColorName): * `Red` - RED
            * `Dark Blue` - DARK_BLUE
            * `Dark Orange` - DARK_ORANGE
            * `Dark Green` - DARK_GREEN
            * `Purple` - PURPLE
            * `Dark Teal` - DARK_TEAL
            * `Pink` - PINK
            * `Orange` - ORANGE
            * `Green` - GREEN
            * `Yellow` - YELLOW
            * `Brown` - BROWN
            * `Dark Red` - DARK_RED
            * `Flat Green` - FLAT_GREEN
            * `Red Orange` - RED_ORANGE
            * `Teal` - TEAL
            * `Light Green` - LIGHT_GREEN
            * `Light Blue` - LIGHT_BLUE
            * `Light Purple` - LIGHT_PURPLE
            * `Light Orange` - LIGHT_ORANGE
            * `Light Pink` - LIGHT_PINK
            * `Tan` - TAN
            * `Dark Gray` - DARK_GRAY
            * `Light Brown` - LIGHT_BROWN
            * `Light Gray` - LIGHT_GRAY
        updated_by_client_duid (Union[Unset, None, str]):
        drafter_duid (Optional[str]):
        recommendation_duid (Optional[str]):
    """

    duid: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    in_trash: bool
    folder_duid: str
    order: str
    title: str
    text: "DocText"
    edited_by_ai: bool
    editor_duids: List[str]
    subscriber_duids: List[str]
    icon_kind: IconKind
    icon_name_or_emoji: str
    color_name: ColorName
    drafter_duid: Optional[str]
    recommendation_duid: Optional[str]
    updated_by_client_duid: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        created_at = self.created_at.isoformat()

        updated_at = self.updated_at.isoformat()

        in_trash = self.in_trash
        folder_duid = self.folder_duid
        order = self.order
        title = self.title
        text = self.text.to_dict()

        edited_by_ai = self.edited_by_ai
        editor_duids = self.editor_duids

        subscriber_duids = self.subscriber_duids

        icon_kind = self.icon_kind.value

        icon_name_or_emoji = self.icon_name_or_emoji
        color_name = self.color_name.value

        updated_by_client_duid = self.updated_by_client_duid
        drafter_duid = self.drafter_duid
        recommendation_duid = self.recommendation_duid

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "createdAt": created_at,
                "updatedAt": updated_at,
                "inTrash": in_trash,
                "folderDuid": folder_duid,
                "order": order,
                "title": title,
                "text": text,
                "editedByAi": edited_by_ai,
                "editorDuids": editor_duids,
                "subscriberDuids": subscriber_duids,
                "iconKind": icon_kind,
                "iconNameOrEmoji": icon_name_or_emoji,
                "colorName": color_name,
                "drafterDuid": drafter_duid,
                "recommendationDuid": recommendation_duid,
            }
        )
        if updated_by_client_duid is not UNSET:
            field_dict["updatedByClientDuid"] = updated_by_client_duid

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.doc_text import DocText

        d = src_dict.copy()
        duid = d.pop("duid")

        created_at = isoparse(d.pop("createdAt"))

        updated_at = isoparse(d.pop("updatedAt"))

        in_trash = d.pop("inTrash")

        folder_duid = d.pop("folderDuid")

        order = d.pop("order")

        title = d.pop("title")

        text = DocText.from_dict(d.pop("text"))

        edited_by_ai = d.pop("editedByAi")

        editor_duids = cast(List[str], d.pop("editorDuids"))

        subscriber_duids = cast(List[str], d.pop("subscriberDuids"))

        icon_kind = IconKind(d.pop("iconKind"))

        icon_name_or_emoji = d.pop("iconNameOrEmoji")

        color_name = ColorName(d.pop("colorName"))

        updated_by_client_duid = d.pop("updatedByClientDuid", UNSET)

        drafter_duid = d.pop("drafterDuid")

        recommendation_duid = d.pop("recommendationDuid")

        doc = cls(
            duid=duid,
            created_at=created_at,
            updated_at=updated_at,
            in_trash=in_trash,
            folder_duid=folder_duid,
            order=order,
            title=title,
            text=text,
            edited_by_ai=edited_by_ai,
            editor_duids=editor_duids,
            subscriber_duids=subscriber_duids,
            icon_kind=icon_kind,
            icon_name_or_emoji=icon_name_or_emoji,
            color_name=color_name,
            updated_by_client_duid=updated_by_client_duid,
            drafter_duid=drafter_duid,
            recommendation_duid=recommendation_duid,
        )

        doc.additional_properties = d
        return doc

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Optional, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.color_name import ColorName

T = TypeVar("T", bound="TaskAttachment")


@_attrs_define
class TaskAttachment:
    """
    Attributes:
        duid (str):
        order (str):
        name (str):
        kind (str):
        file_url (str):
        color_name (ColorName): * `Red` - RED
            * `Dark Blue` - DARK_BLUE
            * `Dark Orange` - DARK_ORANGE
            * `Dark Green` - DARK_GREEN
            * `Purple` - PURPLE
            * `Dark Teal` - DARK_TEAL
            * `Pink` - PINK
            * `Orange` - ORANGE
            * `Green` - GREEN
            * `Yellow` - YELLOW
            * `Brown` - BROWN
            * `Dark Red` - DARK_RED
            * `Flat Green` - FLAT_GREEN
            * `Red Orange` - RED_ORANGE
            * `Teal` - TEAL
            * `Light Green` - LIGHT_GREEN
            * `Light Blue` - LIGHT_BLUE
            * `Light Purple` - LIGHT_PURPLE
            * `Light Orange` - LIGHT_ORANGE
            * `Light Pink` - LIGHT_PINK
            * `Tan` - TAN
            * `Dark Gray` - DARK_GRAY
            * `Light Brown` - LIGHT_BROWN
            * `Light Gray` - LIGHT_GRAY
        recommendation_duid (Optional[str]):
    """

    duid: str
    order: str
    name: str
    kind: str
    file_url: str
    color_name: ColorName
    recommendation_duid: Optional[str]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        order = self.order
        name = self.name
        kind = self.kind
        file_url = self.file_url
        color_name = self.color_name.value

        recommendation_duid = self.recommendation_duid

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "order": order,
                "name": name,
                "kind": kind,
                "fileUrl": file_url,
                "colorName": color_name,
                "recommendationDuid": recommendation_duid,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        order = d.pop("order")

        name = d.pop("name")

        kind = d.pop("kind")

        file_url = d.pop("fileUrl")

        color_name = ColorName(d.pop("colorName"))

        recommendation_duid = d.pop("recommendationDuid")

        task_attachment = cls(
            duid=duid,
            order=order,
            name=name,
            kind=kind,
            file_url=file_url,
            color_name=color_name,
            recommendation_duid=recommendation_duid,
        )

        task_attachment.additional_properties = d
        return task_attachment

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class FilterApplicability(str, Enum):
    ARE_NOT_SET = "are not set"
    CONTAIN = "contain"
    CONTAINS = "contains"
    DONT_INCLUDE = "don't include"
    EXIST = "exist"
    EXISTS = "exists"
    INCLUDE = "include"
    IS = "is"
    IS_AFTER = "is after"
    IS_BEFORE = "is before"
    IS_BETWEEN = "is between"
    IS_NOT = "is not"
    IS_NOT_SET = "is not set"

    def __str__(self) -> str:
        return str(self.value)



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.comment import Comment
    from ..models.comment_reaction import CommentReaction
    from ..models.dartboard import Dartboard
    from ..models.doc import Doc
    from ..models.event import Event
    from ..models.event_subscription import EventSubscription
    from ..models.folder import Folder
    from ..models.layout import Layout
    from ..models.option import Option
    from ..models.property_ import Property
    from ..models.relationship import Relationship
    from ..models.relationship_kind import RelationshipKind
    from ..models.space import Space
    from ..models.status import Status
    from ..models.task import Task
    from ..models.task_attachment import TaskAttachment
    from ..models.task_doc_relationship import TaskDocRelationship
    from ..models.task_link import TaskLink
    from ..models.tenant import Tenant
    from ..models.user import User
    from ..models.user_dartboard_layout import UserDartboardLayout
    from ..models.view import View


T = TypeVar("T", bound="ModelsResponse")


@_attrs_define
class ModelsResponse:
    """This is a helper serializer for OpenAPI schema generation for all available models.
    E.g.:
    {
        "dartboards": [...],
        "layouts": [...],
        "relationships": [...],
        ...
        "views": [...],
    }

        Attributes:
            comments (Union[Unset, List['Comment']]):
            reactions (Union[Unset, List['CommentReaction']]):
            dartboards (Union[Unset, List['Dartboard']]):
            docs (Union[Unset, List['Doc']]):
            events (Union[Unset, List['Event']]):
            folders (Union[Unset, List['Folder']]):
            layouts (Union[Unset, List['Layout']]):
            event_subscriptions (Union[Unset, List['EventSubscription']]):
            options (Union[Unset, List['Option']]):
            properties (Union[Unset, List['Property']]):
            relationships (Union[Unset, List['Relationship']]):
            relationship_kinds (Union[Unset, List['RelationshipKind']]):
            spaces (Union[Unset, List['Space']]):
            statuses (Union[Unset, List['Status']]):
            tasks (Union[Unset, List['Task']]):
            attachments (Union[Unset, List['TaskAttachment']]):
            task_doc_relationships (Union[Unset, List['TaskDocRelationship']]):
            links (Union[Unset, List['TaskLink']]):
            tenants (Union[Unset, List['Tenant']]):
            users (Union[Unset, List['User']]):
            user_dartboard_layouts (Union[Unset, List['UserDartboardLayout']]):
            views (Union[Unset, List['View']]):
    """

    comments: Union[Unset, List["Comment"]] = UNSET
    reactions: Union[Unset, List["CommentReaction"]] = UNSET
    dartboards: Union[Unset, List["Dartboard"]] = UNSET
    docs: Union[Unset, List["Doc"]] = UNSET
    events: Union[Unset, List["Event"]] = UNSET
    folders: Union[Unset, List["Folder"]] = UNSET
    layouts: Union[Unset, List["Layout"]] = UNSET
    event_subscriptions: Union[Unset, List["EventSubscription"]] = UNSET
    options: Union[Unset, List["Option"]] = UNSET
    properties: Union[Unset, List["Property"]] = UNSET
    relationships: Union[Unset, List["Relationship"]] = UNSET
    relationship_kinds: Union[Unset, List["RelationshipKind"]] = UNSET
    spaces: Union[Unset, List["Space"]] = UNSET
    statuses: Union[Unset, List["Status"]] = UNSET
    tasks: Union[Unset, List["Task"]] = UNSET
    attachments: Union[Unset, List["TaskAttachment"]] = UNSET
    task_doc_relationships: Union[Unset, List["TaskDocRelationship"]] = UNSET
    links: Union[Unset, List["TaskLink"]] = UNSET
    tenants: Union[Unset, List["Tenant"]] = UNSET
    users: Union[Unset, List["User"]] = UNSET
    user_dartboard_layouts: Union[Unset, List["UserDartboardLayout"]] = UNSET
    views: Union[Unset, List["View"]] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        comments: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.comments, Unset):
            comments = []
            for comments_item_data in self.comments:
                comments_item = comments_item_data.to_dict()

                comments.append(comments_item)

        reactions: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.reactions, Unset):
            reactions = []
            for reactions_item_data in self.reactions:
                reactions_item = reactions_item_data.to_dict()

                reactions.append(reactions_item)

        dartboards: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.dartboards, Unset):
            dartboards = []
            for dartboards_item_data in self.dartboards:
                dartboards_item = dartboards_item_data.to_dict()

                dartboards.append(dartboards_item)

        docs: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.docs, Unset):
            docs = []
            for docs_item_data in self.docs:
                docs_item = docs_item_data.to_dict()

                docs.append(docs_item)

        events: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.events, Unset):
            events = []
            for events_item_data in self.events:
                events_item = events_item_data.to_dict()

                events.append(events_item)

        folders: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.folders, Unset):
            folders = []
            for folders_item_data in self.folders:
                folders_item = folders_item_data.to_dict()

                folders.append(folders_item)

        layouts: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.layouts, Unset):
            layouts = []
            for layouts_item_data in self.layouts:
                layouts_item = layouts_item_data.to_dict()

                layouts.append(layouts_item)

        event_subscriptions: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.event_subscriptions, Unset):
            event_subscriptions = []
            for event_subscriptions_item_data in self.event_subscriptions:
                event_subscriptions_item = event_subscriptions_item_data.to_dict()

                event_subscriptions.append(event_subscriptions_item)

        options: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.options, Unset):
            options = []
            for options_item_data in self.options:
                options_item = options_item_data.to_dict()

                options.append(options_item)

        properties: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.properties, Unset):
            properties = []
            for properties_item_data in self.properties:
                properties_item = properties_item_data.to_dict()

                properties.append(properties_item)

        relationships: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.relationships, Unset):
            relationships = []
            for relationships_item_data in self.relationships:
                relationships_item = relationships_item_data.to_dict()

                relationships.append(relationships_item)

        relationship_kinds: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.relationship_kinds, Unset):
            relationship_kinds = []
            for relationship_kinds_item_data in self.relationship_kinds:
                relationship_kinds_item = relationship_kinds_item_data.to_dict()

                relationship_kinds.append(relationship_kinds_item)

        spaces: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.spaces, Unset):
            spaces = []
            for spaces_item_data in self.spaces:
                spaces_item = spaces_item_data.to_dict()

                spaces.append(spaces_item)

        statuses: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.statuses, Unset):
            statuses = []
            for statuses_item_data in self.statuses:
                statuses_item = statuses_item_data.to_dict()

                statuses.append(statuses_item)

        tasks: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.tasks, Unset):
            tasks = []
            for tasks_item_data in self.tasks:
                tasks_item = tasks_item_data.to_dict()

                tasks.append(tasks_item)

        attachments: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.attachments, Unset):
            attachments = []
            for attachments_item_data in self.attachments:
                attachments_item = attachments_item_data.to_dict()

                attachments.append(attachments_item)

        task_doc_relationships: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.task_doc_relationships, Unset):
            task_doc_relationships = []
            for task_doc_relationships_item_data in self.task_doc_relationships:
                task_doc_relationships_item = task_doc_relationships_item_data.to_dict()

                task_doc_relationships.append(task_doc_relationships_item)

        links: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.links, Unset):
            links = []
            for links_item_data in self.links:
                links_item = links_item_data.to_dict()

                links.append(links_item)

        tenants: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.tenants, Unset):
            tenants = []
            for tenants_item_data in self.tenants:
                tenants_item = tenants_item_data.to_dict()

                tenants.append(tenants_item)

        users: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.users, Unset):
            users = []
            for users_item_data in self.users:
                users_item = users_item_data.to_dict()

                users.append(users_item)

        user_dartboard_layouts: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.user_dartboard_layouts, Unset):
            user_dartboard_layouts = []
            for user_dartboard_layouts_item_data in self.user_dartboard_layouts:
                user_dartboard_layouts_item = user_dartboard_layouts_item_data.to_dict()

                user_dartboard_layouts.append(user_dartboard_layouts_item)

        views: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.views, Unset):
            views = []
            for views_item_data in self.views:
                views_item = views_item_data.to_dict()

                views.append(views_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if comments is not UNSET:
            field_dict["comments"] = comments
        if reactions is not UNSET:
            field_dict["reactions"] = reactions
        if dartboards is not UNSET:
            field_dict["dartboards"] = dartboards
        if docs is not UNSET:
            field_dict["docs"] = docs
        if events is not UNSET:
            field_dict["events"] = events
        if folders is not UNSET:
            field_dict["folders"] = folders
        if layouts is not UNSET:
            field_dict["layouts"] = layouts
        if event_subscriptions is not UNSET:
            field_dict["eventSubscriptions"] = event_subscriptions
        if options is not UNSET:
            field_dict["options"] = options
        if properties is not UNSET:
            field_dict["properties"] = properties
        if relationships is not UNSET:
            field_dict["relationships"] = relationships
        if relationship_kinds is not UNSET:
            field_dict["relationshipKinds"] = relationship_kinds
        if spaces is not UNSET:
            field_dict["spaces"] = spaces
        if statuses is not UNSET:
            field_dict["statuses"] = statuses
        if tasks is not UNSET:
            field_dict["tasks"] = tasks
        if attachments is not UNSET:
            field_dict["attachments"] = attachments
        if task_doc_relationships is not UNSET:
            field_dict["taskDocRelationships"] = task_doc_relationships
        if links is not UNSET:
            field_dict["links"] = links
        if tenants is not UNSET:
            field_dict["tenants"] = tenants
        if users is not UNSET:
            field_dict["users"] = users
        if user_dartboard_layouts is not UNSET:
            field_dict["userDartboardLayouts"] = user_dartboard_layouts
        if views is not UNSET:
            field_dict["views"] = views

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.comment import Comment
        from ..models.comment_reaction import CommentReaction
        from ..models.dartboard import Dartboard
        from ..models.doc import Doc
        from ..models.event import Event
        from ..models.event_subscription import EventSubscription
        from ..models.folder import Folder
        from ..models.layout import Layout
        from ..models.option import Option
        from ..models.property_ import Property
        from ..models.relationship import Relationship
        from ..models.relationship_kind import RelationshipKind
        from ..models.space import Space
        from ..models.status import Status
        from ..models.task import Task
        from ..models.task_attachment import TaskAttachment
        from ..models.task_doc_relationship import TaskDocRelationship
        from ..models.task_link import TaskLink
        from ..models.tenant import Tenant
        from ..models.user import User
        from ..models.user_dartboard_layout import UserDartboardLayout
        from ..models.view import View

        d = src_dict.copy()
        comments = []
        _comments = d.pop("comments", UNSET)
        for comments_item_data in _comments or []:
            comments_item = Comment.from_dict(comments_item_data)

            comments.append(comments_item)

        reactions = []
        _reactions = d.pop("reactions", UNSET)
        for reactions_item_data in _reactions or []:
            reactions_item = CommentReaction.from_dict(reactions_item_data)

            reactions.append(reactions_item)

        dartboards = []
        _dartboards = d.pop("dartboards", UNSET)
        for dartboards_item_data in _dartboards or []:
            dartboards_item = Dartboard.from_dict(dartboards_item_data)

            dartboards.append(dartboards_item)

        docs = []
        _docs = d.pop("docs", UNSET)
        for docs_item_data in _docs or []:
            docs_item = Doc.from_dict(docs_item_data)

            docs.append(docs_item)

        events = []
        _events = d.pop("events", UNSET)
        for events_item_data in _events or []:
            events_item = Event.from_dict(events_item_data)

            events.append(events_item)

        folders = []
        _folders = d.pop("folders", UNSET)
        for folders_item_data in _folders or []:
            folders_item = Folder.from_dict(folders_item_data)

            folders.append(folders_item)

        layouts = []
        _layouts = d.pop("layouts", UNSET)
        for layouts_item_data in _layouts or []:
            layouts_item = Layout.from_dict(layouts_item_data)

            layouts.append(layouts_item)

        event_subscriptions = []
        _event_subscriptions = d.pop("eventSubscriptions", UNSET)
        for event_subscriptions_item_data in _event_subscriptions or []:
            event_subscriptions_item = EventSubscription.from_dict(event_subscriptions_item_data)

            event_subscriptions.append(event_subscriptions_item)

        options = []
        _options = d.pop("options", UNSET)
        for options_item_data in _options or []:
            options_item = Option.from_dict(options_item_data)

            options.append(options_item)

        properties = []
        _properties = d.pop("properties", UNSET)
        for properties_item_data in _properties or []:
            properties_item = Property.from_dict(properties_item_data)

            properties.append(properties_item)

        relationships = []
        _relationships = d.pop("relationships", UNSET)
        for relationships_item_data in _relationships or []:
            relationships_item = Relationship.from_dict(relationships_item_data)

            relationships.append(relationships_item)

        relationship_kinds = []
        _relationship_kinds = d.pop("relationshipKinds", UNSET)
        for relationship_kinds_item_data in _relationship_kinds or []:
            relationship_kinds_item = RelationshipKind.from_dict(relationship_kinds_item_data)

            relationship_kinds.append(relationship_kinds_item)

        spaces = []
        _spaces = d.pop("spaces", UNSET)
        for spaces_item_data in _spaces or []:
            spaces_item = Space.from_dict(spaces_item_data)

            spaces.append(spaces_item)

        statuses = []
        _statuses = d.pop("statuses", UNSET)
        for statuses_item_data in _statuses or []:
            statuses_item = Status.from_dict(statuses_item_data)

            statuses.append(statuses_item)

        tasks = []
        _tasks = d.pop("tasks", UNSET)
        for tasks_item_data in _tasks or []:
            tasks_item = Task.from_dict(tasks_item_data)

            tasks.append(tasks_item)

        attachments = []
        _attachments = d.pop("attachments", UNSET)
        for attachments_item_data in _attachments or []:
            attachments_item = TaskAttachment.from_dict(attachments_item_data)

            attachments.append(attachments_item)

        task_doc_relationships = []
        _task_doc_relationships = d.pop("taskDocRelationships", UNSET)
        for task_doc_relationships_item_data in _task_doc_relationships or []:
            task_doc_relationships_item = TaskDocRelationship.from_dict(task_doc_relationships_item_data)

            task_doc_relationships.append(task_doc_relationships_item)

        links = []
        _links = d.pop("links", UNSET)
        for links_item_data in _links or []:
            links_item = TaskLink.from_dict(links_item_data)

            links.append(links_item)

        tenants = []
        _tenants = d.pop("tenants", UNSET)
        for tenants_item_data in _tenants or []:
            tenants_item = Tenant.from_dict(tenants_item_data)

            tenants.append(tenants_item)

        users = []
        _users = d.pop("users", UNSET)
        for users_item_data in _users or []:
            users_item = User.from_dict(users_item_data)

            users.append(users_item)

        user_dartboard_layouts = []
        _user_dartboard_layouts = d.pop("userDartboardLayouts", UNSET)
        for user_dartboard_layouts_item_data in _user_dartboard_layouts or []:
            user_dartboard_layouts_item = UserDartboardLayout.from_dict(user_dartboard_layouts_item_data)

            user_dartboard_layouts.append(user_dartboard_layouts_item)

        views = []
        _views = d.pop("views", UNSET)
        for views_item_data in _views or []:
            views_item = View.from_dict(views_item_data)

            views.append(views_item)

        models_response = cls(
            comments=comments,
            reactions=reactions,
            dartboards=dartboards,
            docs=docs,
            events=events,
            folders=folders,
            layouts=layouts,
            event_subscriptions=event_subscriptions,
            options=options,
            properties=properties,
            relationships=relationships,
            relationship_kinds=relationship_kinds,
            spaces=spaces,
            statuses=statuses,
            tasks=tasks,
            attachments=attachments,
            task_doc_relationships=task_doc_relationships,
            links=links,
            tenants=tenants,
            users=users,
            user_dartboard_layouts=user_dartboard_layouts,
            views=views,
        )

        models_response.additional_properties = d
        return models_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.relationship_kind_kind import RelationshipKindKind
from ..types import UNSET, Unset

T = TypeVar("T", bound="RelationshipKindCreate")


@_attrs_define
class RelationshipKindCreate:
    """
    Attributes:
        duid (str):
        directed (bool):
        forward_text (str):
        kind (Union[Unset, RelationshipKindKind]): * `Parent Of` - PARENT_OF
            * `Blocks` - BLOCKS
            * `Relates To` - RELATES_TO
            * `Duplicates` - DUPLICATES
            * `Custom` - CUSTOM
        backward_text (Union[Unset, None, str]):
    """

    duid: str
    directed: bool
    forward_text: str
    kind: Union[Unset, RelationshipKindKind] = UNSET
    backward_text: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        directed = self.directed
        forward_text = self.forward_text
        kind: Union[Unset, str] = UNSET
        if not isinstance(self.kind, Unset):
            kind = self.kind.value

        backward_text = self.backward_text

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "directed": directed,
                "forwardText": forward_text,
            }
        )
        if kind is not UNSET:
            field_dict["kind"] = kind
        if backward_text is not UNSET:
            field_dict["backwardText"] = backward_text

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        directed = d.pop("directed")

        forward_text = d.pop("forwardText")

        _kind = d.pop("kind", UNSET)
        kind: Union[Unset, RelationshipKindKind]
        if isinstance(_kind, Unset):
            kind = UNSET
        else:
            kind = RelationshipKindKind(_kind)

        backward_text = d.pop("backwardText", UNSET)

        relationship_kind_create = cls(
            duid=duid,
            directed=directed,
            forward_text=forward_text,
            kind=kind,
            backward_text=backward_text,
        )

        relationship_kind_create.additional_properties = d
        return relationship_kind_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

if TYPE_CHECKING:
    from ..models.task_notion_document_block_children_map import TaskNotionDocumentBlockChildrenMap
    from ..models.task_notion_document_block_map import TaskNotionDocumentBlockMap
    from ..models.task_notion_document_page_map import TaskNotionDocumentPageMap


T = TypeVar("T", bound="TaskNotionDocument")


@_attrs_define
class TaskNotionDocument:
    """
    Attributes:
        page_id (str):
        refreshing (bool):
        exists_and_access_granted (Optional[bool]):
        last_refresh_at (Optional[datetime.datetime]):
        page_map (Optional[TaskNotionDocumentPageMap]):
        block_map (Optional[TaskNotionDocumentBlockMap]):
        block_children_map (Optional[TaskNotionDocumentBlockChildrenMap]):
    """

    page_id: str
    refreshing: bool
    exists_and_access_granted: Optional[bool]
    last_refresh_at: Optional[datetime.datetime]
    page_map: Optional["TaskNotionDocumentPageMap"]
    block_map: Optional["TaskNotionDocumentBlockMap"]
    block_children_map: Optional["TaskNotionDocumentBlockChildrenMap"]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        page_id = self.page_id
        refreshing = self.refreshing
        exists_and_access_granted = self.exists_and_access_granted
        last_refresh_at = self.last_refresh_at.isoformat() if self.last_refresh_at else None

        page_map = self.page_map.to_dict() if self.page_map else None

        block_map = self.block_map.to_dict() if self.block_map else None

        block_children_map = self.block_children_map.to_dict() if self.block_children_map else None

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "pageId": page_id,
                "refreshing": refreshing,
                "existsAndAccessGranted": exists_and_access_granted,
                "lastRefreshAt": last_refresh_at,
                "pageMap": page_map,
                "blockMap": block_map,
                "blockChildrenMap": block_children_map,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.task_notion_document_block_children_map import TaskNotionDocumentBlockChildrenMap
        from ..models.task_notion_document_block_map import TaskNotionDocumentBlockMap
        from ..models.task_notion_document_page_map import TaskNotionDocumentPageMap

        d = src_dict.copy()
        page_id = d.pop("pageId")

        refreshing = d.pop("refreshing")

        exists_and_access_granted = d.pop("existsAndAccessGranted")

        _last_refresh_at = d.pop("lastRefreshAt")
        last_refresh_at: Optional[datetime.datetime]
        if _last_refresh_at is None:
            last_refresh_at = None
        else:
            last_refresh_at = isoparse(_last_refresh_at)

        _page_map = d.pop("pageMap")
        page_map: Optional[TaskNotionDocumentPageMap]
        if _page_map is None:
            page_map = None
        else:
            page_map = TaskNotionDocumentPageMap.from_dict(_page_map)

        _block_map = d.pop("blockMap")
        block_map: Optional[TaskNotionDocumentBlockMap]
        if _block_map is None:
            block_map = None
        else:
            block_map = TaskNotionDocumentBlockMap.from_dict(_block_map)

        _block_children_map = d.pop("blockChildrenMap")
        block_children_map: Optional[TaskNotionDocumentBlockChildrenMap]
        if _block_children_map is None:
            block_children_map = None
        else:
            block_children_map = TaskNotionDocumentBlockChildrenMap.from_dict(_block_children_map)

        task_notion_document = cls(
            page_id=page_id,
            refreshing=refreshing,
            exists_and_access_granted=exists_and_access_granted,
            last_refresh_at=last_refresh_at,
            page_map=page_map,
            block_map=block_map,
            block_children_map=block_children_map,
        )

        task_notion_document.additional_properties = d
        return task_notion_document

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.color_name import ColorName
from ..models.status_kind import StatusKind
from ..types import UNSET, Unset

T = TypeVar("T", bound="StatusCreate")


@_attrs_define
class StatusCreate:
    """
    Attributes:
        duid (str):
        property_duid (str):
        kind (StatusKind): * `Unstarted` - UNSTARTED
            * `Started` - STARTED
            * `Blocked` - BLOCKED
            * `Finished` - FINISHED
            * `Canceled` - CANCELED
        order (str):
        title (Union[Unset, str]):
        color_name (Union[Unset, ColorName]): * `Red` - RED
            * `Dark Blue` - DARK_BLUE
            * `Dark Orange` - DARK_ORANGE
            * `Dark Green` - DARK_GREEN
            * `Purple` - PURPLE
            * `Dark Teal` - DARK_TEAL
            * `Pink` - PINK
            * `Orange` - ORANGE
            * `Green` - GREEN
            * `Yellow` - YELLOW
            * `Brown` - BROWN
            * `Dark Red` - DARK_RED
            * `Flat Green` - FLAT_GREEN
            * `Red Orange` - RED_ORANGE
            * `Teal` - TEAL
            * `Light Green` - LIGHT_GREEN
            * `Light Blue` - LIGHT_BLUE
            * `Light Purple` - LIGHT_PURPLE
            * `Light Orange` - LIGHT_ORANGE
            * `Light Pink` - LIGHT_PINK
            * `Tan` - TAN
            * `Dark Gray` - DARK_GRAY
            * `Light Brown` - LIGHT_BROWN
            * `Light Gray` - LIGHT_GRAY
        description (Union[Unset, str]):
    """

    duid: str
    property_duid: str
    kind: StatusKind
    order: str
    title: Union[Unset, str] = UNSET
    color_name: Union[Unset, ColorName] = UNSET
    description: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        property_duid = self.property_duid
        kind = self.kind.value

        order = self.order
        title = self.title
        color_name: Union[Unset, str] = UNSET
        if not isinstance(self.color_name, Unset):
            color_name = self.color_name.value

        description = self.description

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "propertyDuid": property_duid,
                "kind": kind,
                "order": order,
            }
        )
        if title is not UNSET:
            field_dict["title"] = title
        if color_name is not UNSET:
            field_dict["colorName"] = color_name
        if description is not UNSET:
            field_dict["description"] = description

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        property_duid = d.pop("propertyDuid")

        kind = StatusKind(d.pop("kind"))

        order = d.pop("order")

        title = d.pop("title", UNSET)

        _color_name = d.pop("colorName", UNSET)
        color_name: Union[Unset, ColorName]
        if isinstance(_color_name, Unset):
            color_name = UNSET
        else:
            color_name = ColorName(_color_name)

        description = d.pop("description", UNSET)

        status_create = cls(
            duid=duid,
            property_duid=property_duid,
            kind=kind,
            order=order,
            title=title,
            color_name=color_name,
            description=description,
        )

        status_create.additional_properties = d
        return status_create

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.property_kind import PropertyKind
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.property_adtl import PropertyAdtl


T = TypeVar("T", bound="Property")


@_attrs_define
class Property:
    """
    Attributes:
        duid (str):
        kind (PropertyKind): * `Default: Status` - DEFAULT_STATUS
            * `Default: Assignee` - DEFAULT_ASSIGNEE
            * `Default: Dates` - DEFAULT_DATES
            * `Default: Priority` - DEFAULT_PRIORITY
            * `Default: Tags` - DEFAULT_TAGS
            * `Default: Size` - DEFAULT_SIZE
            * `Default: Created` - DEFAULT_CREATED_AT
            * `Default: Created by` - DEFAULT_CREATED_BY
            * `Default: Last updated` - DEFAULT_UPDATED_AT
            * `Default: Last updated by` - DEFAULT_UPDATED_BY
            * `Text` - TEXT
            * `Number` - NUMBER
            * `Checkbox` - CHECKBOX
            * `Select` - SELECT
            * `Multiselect` - MULTISELECT
            * `Status` - STATUS
            * `User` - USER
            * `Dates` - DATES
        order (str):
        hidden (bool):
        title (str):
        description (str):
        adtl (PropertyAdtl):
        updated_by_client_duid (Union[Unset, None, str]):
    """

    duid: str
    kind: PropertyKind
    order: str
    hidden: bool
    title: str
    description: str
    adtl: "PropertyAdtl"
    updated_by_client_duid: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        kind = self.kind.value

        order = self.order
        hidden = self.hidden
        title = self.title
        description = self.description
        adtl = self.adtl.to_dict()

        updated_by_client_duid = self.updated_by_client_duid

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "kind": kind,
                "order": order,
                "hidden": hidden,
                "title": title,
                "description": description,
                "adtl": adtl,
            }
        )
        if updated_by_client_duid is not UNSET:
            field_dict["updatedByClientDuid"] = updated_by_client_duid

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.property_adtl import PropertyAdtl

        d = src_dict.copy()
        duid = d.pop("duid")

        kind = PropertyKind(d.pop("kind"))

        order = d.pop("order")

        hidden = d.pop("hidden")

        title = d.pop("title")

        description = d.pop("description")

        adtl = PropertyAdtl.from_dict(d.pop("adtl"))

        updated_by_client_duid = d.pop("updatedByClientDuid", UNSET)

        property_ = cls(
            duid=duid,
            kind=kind,
            order=order,
            hidden=hidden,
            title=title,
            description=description,
            adtl=adtl,
            updated_by_client_duid=updated_by_client_duid,
        )

        property_.additional_properties = d
        return property_

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import TYPE_CHECKING, Any, Dict, List, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.models_response import ModelsResponse


T = TypeVar("T", bound="TransactionResponse")


@_attrs_define
class TransactionResponse:
    """
    Attributes:
        duid (str):
        success (bool):
        message (str):
        models (ModelsResponse): This is a helper serializer for OpenAPI schema generation for all available models.
            E.g.:
            {
                "dartboards": [...],
                "layouts": [...],
                "relationships": [...],
                ...
                "views": [...],
            }
    """

    duid: str
    success: bool
    message: str
    models: "ModelsResponse"
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        success = self.success
        message = self.message
        models = self.models.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
                "success": success,
                "message": message,
                "models": models,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        from ..models.models_response import ModelsResponse

        d = src_dict.copy()
        duid = d.pop("duid")

        success = d.pop("success")

        message = d.pop("message")

        models = ModelsResponse.from_dict(d.pop("models"))

        transaction_response = cls(
            duid=duid,
            success=success,
            message=message,
            models=models,
        )

        transaction_response.additional_properties = d
        return transaction_response

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ColorName(str, Enum):
    BROWN = "Brown"
    DARK_BLUE = "Dark Blue"
    DARK_GRAY = "Dark Gray"
    DARK_GREEN = "Dark Green"
    DARK_ORANGE = "Dark Orange"
    DARK_RED = "Dark Red"
    DARK_TEAL = "Dark Teal"
    FLAT_GREEN = "Flat Green"
    GREEN = "Green"
    LIGHT_BLUE = "Light Blue"
    LIGHT_BROWN = "Light Brown"
    LIGHT_GRAY = "Light Gray"
    LIGHT_GREEN = "Light Green"
    LIGHT_ORANGE = "Light Orange"
    LIGHT_PINK = "Light Pink"
    LIGHT_PURPLE = "Light Purple"
    ORANGE = "Orange"
    PINK = "Pink"
    PURPLE = "Purple"
    RED = "Red"
    RED_ORANGE = "Red Orange"
    TAN = "Tan"
    TEAL = "Teal"
    YELLOW = "Yellow"

    def __str__(self) -> str:
        return str(self.value)



import datetime
from typing import Any, Dict, List, Optional, Type, TypeVar

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.batch import Batch

T = TypeVar("T", bound="YcIntegration")


@_attrs_define
class YcIntegration:
    """
    Attributes:
        batch (Batch): * `S05` - S05
            * `W06` - W06
            * `S06` - S06
            * `W07` - W07
            * `S07` - S07
            * `W08` - W08
            * `S08` - S08
            * `W09` - W09
            * `S09` - S09
            * `W10` - W10
            * `S10` - S10
            * `W11` - W11
            * `S11` - S11
            * `W12` - W12
            * `S12` - S12
            * `W13` - W13
            * `S13` - S13
            * `W14` - W14
            * `S14` - S14
            * `W15` - W15
            * `S15` - S15
            * `W16` - W16
            * `S16` - S16
            * `IK12` - IK12
            * `W17` - W17
            * `S17` - S17
            * `W18` - W18
            * `S18` - S18
            * `W19` - W19
            * `S19` - S19
            * `W20` - W20
            * `S20` - S20
            * `W21` - W21
            * `S21` - S21
            * `W22` - W22
            * `S22` - S22
            * `W23` - W23
        batch_full_name (str):
        demo_day_date_str (Optional[datetime.date]):
    """

    batch: Batch
    batch_full_name: str
    demo_day_date_str: Optional[datetime.date]
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        batch = self.batch.value

        batch_full_name = self.batch_full_name
        demo_day_date_str = self.demo_day_date_str.isoformat() if self.demo_day_date_str else None

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "batch": batch,
                "batchFullName": batch_full_name,
                "demoDayDateStr": demo_day_date_str,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        batch = Batch(d.pop("batch"))

        batch_full_name = d.pop("batchFullName")

        _demo_day_date_str = d.pop("demoDayDateStr")
        demo_day_date_str: Optional[datetime.date]
        if _demo_day_date_str is None:
            demo_day_date_str = None
        else:
            demo_day_date_str = isoparse(_demo_day_date_str).date()

        yc_integration = cls(
            batch=batch,
            batch_full_name=batch_full_name,
            demo_day_date_str=demo_day_date_str,
        )

        yc_integration.additional_properties = d
        return yc_integration

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any, Dict, List, Type, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.task_link_kind import TaskLinkKind
from ..types import UNSET, Unset

T = TypeVar("T", bound="TaskLinkUpdate")


@_attrs_define
class TaskLinkUpdate:
    """
    Attributes:
        duid (str):
        task_duid (Union[Unset, str]):
        order (Union[Unset, str]):
        kind (Union[Unset, TaskLinkKind]): * `Standard` - STANDARD
            * `Source From Template` - SOURCE_FROM_TEMPLATE
            * `Source From Third Party` - SOURCE_FROM_THIRD_PARTY
            * `GitHub Branch` - GITHUB_BRANCH
            * `GitHub Pull Request` - GITHUB_PULL_REQUEST
            * `GitHub Expansion` - GITHUB_EXPANSION
            * `Notion Link` - NOTION_LINK
            * `Notion Document` - NOTION_DOCUMENT
            * `Notion Document Doesnt Exist` - NOTION_DOCUMENT_DOESNT_EXIST
            * `Notion Document Parse Failed` - NOTION_DOCUMENT_PARSE_FAILED
            * `Notion Expansion` - NOTION_EXPANSION
            * `Slack Expansion` - SLACK_EXPANSION
        url (Union[Unset, str]):
        title (Union[Unset, None, str]):
        icon_url (Union[Unset, None, str]):
    """

    duid: str
    task_duid: Union[Unset, str] = UNSET
    order: Union[Unset, str] = UNSET
    kind: Union[Unset, TaskLinkKind] = UNSET
    url: Union[Unset, str] = UNSET
    title: Union[Unset, None, str] = UNSET
    icon_url: Union[Unset, None, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        duid = self.duid
        task_duid = self.task_duid
        order = self.order
        kind: Union[Unset, str] = UNSET
        if not isinstance(self.kind, Unset):
            kind = self.kind.value

        url = self.url
        title = self.title
        icon_url = self.icon_url

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "duid": duid,
            }
        )
        if task_duid is not UNSET:
            field_dict["taskDuid"] = task_duid
        if order is not UNSET:
            field_dict["order"] = order
        if kind is not UNSET:
            field_dict["kind"] = kind
        if url is not UNSET:
            field_dict["url"] = url
        if title is not UNSET:
            field_dict["title"] = title
        if icon_url is not UNSET:
            field_dict["iconUrl"] = icon_url

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        d = src_dict.copy()
        duid = d.pop("duid")

        task_duid = d.pop("taskDuid", UNSET)

        order = d.pop("order", UNSET)

        _kind = d.pop("kind", UNSET)
        kind: Union[Unset, TaskLinkKind]
        if isinstance(_kind, Unset):
            kind = UNSET
        else:
            kind = TaskLinkKind(_kind)

        url = d.pop("url", UNSET)

        title = d.pop("title", UNSET)

        icon_url = d.pop("iconUrl", UNSET)

        task_link_update = cls(
            duid=duid,
            task_duid=task_duid,
            order=order,
            kind=kind,
            url=url,
            title=title,
            icon_url=icon_url,
        )

        task_link_update.additional_properties = d
        return task_link_update

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
